import re
import sys
import time
import numpy as np
import os
import warnings

from .utils import (path_depth, gen_hparams_string, beam_algorithm_generator, default_runner, run_worker,
                    build_device_list)
import torch
import copy
import pandas as pd
import torch.multiprocessing as mp
from functools import partial


from ..utils import (set_seed, find_free_port, check_if_port_is_available, is_notebook,
                     find_port, as_numpy, lazy_property, check_type, beam_device, beam_service_port)
from ..path import beam_path, BeamPath, beam_key
from ..logger import beam_logger as logger
from ..config import print_beam_hyperparameters, BeamConfig, UniversalConfig

warnings.filterwarnings('ignore', category=FutureWarning)


class Experiment(object):
    """
    Experiment name:
    <algorithm name>_<identifier>_exp_<number>_<time>


    Experiment number and overriding experiments

    These parameters are responsible for which experiment to load or to generate:
    the name of the experiment is <alg>_<identifier>_exp_<num>_<time>
    The possible configurations:
    reload = False, override = True: always overrides last experiment (default configuration)
    reload = False, override = False: always append experiment to the list (increment experiment num)
    reload = True, resume = -1: resume to the last experiment
    reload = True, resume = <n>: resume to the <n> experiment


    :param args:
    """

    def __init__(self, args, hpo=None, trial=None, print_hyperparameters=None, reload_iloc=None,
                 reload_dir=None, reload_loc=None, reload_name=None):
        """

        @param args:
        @param hpo:
        @param trial:
        @param print_hyperparameters: If None, default behavior is to print hyperparameters only outside of jupyter notebooks
        """

        if print_hyperparameters is None:
            print_hyperparameters = not is_notebook()
        self.print_hyperparameters = print_hyperparameters

        self.tensorboard_hparams = {}

        if not isinstance(args, BeamConfig):
            args = BeamConfig(config=args)

        self.vars_args = dict(args.items())
        for k, v in self.vars_args.items():
            param_type = check_type(v)
            if param_type.major == 'scalar' and param_type.element in ['bool', 'str', 'int', 'float']:
                self.tensorboard_hparams[k] = v

        self.hparams = copy.deepcopy(args)
        logger


        set_seed(seed=self.hparams.seed, constant=0, increment=False, deterministic=self.hparams.deterministic)

        # parameters
        self.start_time = time.time()
        self.exptime = time.strftime("%Y%m%d_%H%M%S", time.localtime())
        self.device = beam_device(self.hparams.device)
        self.device_list = None

        self.exp_name = None
        self.load_model = False

        if reload_dir is None:
            root_path = beam_path(self.hparams.logs_path)
            base_dir = root_path.joinpath(self.hparams.project_name, self.hparams.algorithm, self.hparams.identifier)

            pattern = re.compile("\A\d{6}_\d{8}_\d{6}\Z")

            if base_dir.exists():
                assert base_dir.is_dir(), f"Experiment directory contains an existing file: {base_dir}"
                exp_names = list(filter(lambda x: re.match(pattern, x.name) is not None, base_dir.iterdir()))
                exp_indices = np.array([int(d.name.split('_')[0]) for d in exp_names])
            else:
                exp_names = []
                exp_indices = np.array([])

            exp_num = None
            if self.hparams.reload:

                if type(self.hparams.resume) is str:

                    if base_dir.joinpath(self.hparams.resume).is_dir():
                        self.exp_name = self.hparams.resume
                        exp_num = int(self.exp_name.split('_')[0])
                        self.load_model = True

                elif self.hparams.resume >= 0:
                    ind = np.nonzero(exp_indices == self.hparams.resume)[0]
                    if len(ind):
                        self.exp_name = exp_names[ind[0]]
                        exp_num = self.hparams.resume
                        self.load_model = True

                else:
                    if len(exp_indices):
                        ind = np.argmax(exp_indices)
                        self.exp_name = exp_names[ind]
                        exp_num = exp_indices[ind]
                        self.load_model = True

            else:

                if self.hparams.override and len(exp_indices):

                    ind = np.argmax(exp_indices)
                    self.exp_name = exp_names[ind]
                    exp_num = exp_indices[ind]
                else:
                    self.hparams.set('override', False)

            if self.hparams.reload and not self.load_model:
                logger.warning(f"Did not find existing experiment to match your specifications: basedir={base_dir} resume={self.hparams.resume}")

            if self.exp_name is None:
                exp_num = np.max(exp_indices) + 1 if len(exp_indices) else 0
                self.exp_name = "%06d_%s" % (exp_num, self.exptime)

            self.exp_num = exp_num
            # init experiment parameters
            self.experiment_dir = base_dir.joinpath(self.exp_name)

        else:
            self.experiment_dir = beam_path(reload_dir)
            self.exp_name = self.experiment_dir.name
            self.exp_num = None
            self.load_model = True

        # set dirs
        self.tensorboard_dir = self.experiment_dir.joinpath('tensorboard')
        self.checkpoints_dir = self.experiment_dir.joinpath('checkpoints')
        self.results_dir = self.experiment_dir.joinpath('results')
        self.code_dir = self.experiment_dir.joinpath('code')

        self.store_init_path = None
        if self.hparams.get('store_init_args', default=True):
            self.store_init_path = self.experiment_dir.joinpath('init_alg_args.pkl')

        if self.load_model:
            logger.cleanup()
            logger.add_file_handlers(self.experiment_dir.joinpath('experiment.log'))
            logger.info(f"Resuming existing experiment")

        self.tensorboard_writer = None

        self.rank = 0
        self.world_size = args.n_gpus
        if hasattr(args, 'n_gpus_per_worker') and args.n_gpus_per_worker is not None:
            self.world_size = self.world_size // args.n_gpus_per_worker
            assert self.world_size * args.n_gpus_per_worker == args.n_gpus, \
                (f"Total number of gpus ({args.n_gpus}) is not divisible by number of gpus per worker "
                 f"({args.n_gpus_per_worker})")

        if self.world_size > 1:
            torch.multiprocessing.set_sharing_strategy('file_system')

        # replace zero split_dataset_seed to none (non-deterministic split) - if zero
        if self.hparams.split_dataset_seed == 0:
            self.hparams.set('split_dataset_seed', None)

        # fill the batch size

        if self.hparams.batch_size_train is None:
            self.hparams.set('batch_size_train', self.hparams.batch_size)

        if self.hparams.batch_size_eval is None:
            self.hparams.set('batch_size_eval', self.hparams.batch_size)

        if self.hparams.batch_size is None:
            self.hparams.set('batch_size', self.hparams.batch_size_train)

        # build the hyperparamter class which will be sent to the dataset and algorithm classes

        if self.load_model:

            if reload_iloc is None and reload_loc is None and reload_name is not None:

                if self.hparams.reload_checkpoint == 'last':
                    reload_iloc, reload_loc, reload_name = -1, None, None
                elif self.hparams.reload_checkpoint != 'best':
                    try:
                        reload_iloc, reload_loc, reload_name = None, int(self.hparams.reload_checkpoint), None
                    except ValueError:
                        reload_iloc, reload_loc, reload_name = None, None, self.hparams.reload_checkpoint

            reload_path = self.reload_checkpoint(iloc=reload_iloc, loc=reload_loc, name=reload_name)
        else:
            reload_path = None

        self.hparams.set('reload_path', reload_path)

        self.trial = trial

        self.hpo = hpo
        self.distributed_training = False

        if self.device.type == 'cuda':
            self.device_list = build_device_list(self.hparams)

        self.comet_exp = None
        self.comet_writer = None
        self.mlflow_writer = None
        self.logs_path_is_built = False
        self.source_dir = None

    def algorithm_generator(self, *args, **kwargs):
        return beam_algorithm_generator(self, *args, **kwargs)

    @lazy_property
    def training_framework(self):
        return self.hparams.get('training_framework', 'torch')

    @lazy_property
    def llm(self):
        if self.hparams.llm is not None:
            from ..llm import beam_llm
            return beam_llm(self.hparams.llm)
        return None

    @staticmethod
    def reload_from_path(path, override_hparams=None, reload_iloc=None, reload_loc=None, reload_name=None, **argv):

        path = beam_path(path)
        logger.info(f"Reload experiment from path: {path}")

        args = BeamConfig.from_path(path.joinpath('args.pkl'))
        args.override = False
        args.reload = True

        if override_hparams is not None:
            for k, v in override_hparams.items():
                if k in ['reload', 'resume', 'override', 'project', 'algorithm', 'identifier', 'logs_path']:
                    continue
                setattr(args, k, v)

        return Experiment(args, reload_iloc=reload_iloc, reload_loc=reload_loc, reload_name=reload_name,
                          reload_dir=path, **argv)

    def reload_best_checkpoint(self, alg=None):
            return self.reload_checkpoint(alg, name='checkpoint_best')

    def reload_checkpoint(self, alg=None, iloc=None, loc=None, name=None):

        if iloc is None and loc is None and name is None:
            if self.checkpoints_dir.joinpath('checkpoint_best').is_file():
                name = 'checkpoint_best'
            else:
                iloc = -1

        if name is not None:
            path = self.checkpoints_dir.joinpath(name)

        else:

            checkpoints = list(self.checkpoints_dir.iterdir())
            checkpoints = [c for c in checkpoints if str(c).split('_')[-1].isnumeric()]
            checkpoints_int = [int(c.name.split('_')[-1]) for c in checkpoints]

            if not(len(checkpoints)):
                logger.error(f"Directory of checkpoints does not contain valid checkpoint files")
                return

            checkpoints = pd.DataFrame({'name': checkpoints}, index=checkpoints_int)
            checkpoints = checkpoints.sort_index()

            if loc is not None:
                chp = str(checkpoints.loc[loc]['name'])
                path = self.checkpoints_dir.joinpath(chp)
            else:
                chp = str(checkpoints.iloc[iloc]['name'])
                path = self.checkpoints_dir.joinpath(chp)

        logger.info(f"Reload experiment from checkpoint: {path.name}")

        if alg is not None:
            alg.load_checkpoint(path, hparams=False)
            return alg
        else:
            return path

    def get_alg_init_args(self):

        try:
            init_args = self.store_init_path.read(ext='.pkl')
        except FileNotFoundError:
            return [], {}

        return init_args['args'], init_args['kwargs']

    def set_rank(self, rank, world_size, devices=None):

        self.rank = rank
        self.world_size = world_size

        self.distributed_training = self.world_size > 1
        self.hparams.set('enable_tqdm', self.hparams.enable_tqdm and (rank == 0))

        if self.device.type != 'cpu' and world_size > 1:
            if devices is None:
                self.device = beam_device(self.device_list[rank])
            else:
                self.device = beam_device(devices[0])
                self.device_list = devices

        logger.info(f'Worker {rank + 1} will be running on device={str(self.device)}')

    def writer_control(self, networks=None, inputs=None):

        if self.tensorboard_writer is None and self.hparams.tensorboard:
            if isinstance(self.tensorboard_dir, BeamPath):
                from tensorboardX import SummaryWriter
                self.tensorboard_writer = SummaryWriter(log_dir=str(self.tensorboard_dir.joinpath('logs')),
                                                        comment=self.hparams.identifier)
            else:
                logger.warning(f"Tensorboard directory is not a BeamPath object. Tensorboard will not be enabled.")

        if self.hparams.comet:

            api_key = self.hparams.COMET_API_KEY
            if api_key is None:
                # api_key = os.environ.get('COMET_API_KEY', None)
                api_key = beam_key('COMET_API_KEY')
            git_directory = self.hparams.git_directory
            if git_directory is None and isinstance(self.code_dir, BeamPath):
                git_directory = str(self.code_dir)
            if git_directory is not None:
                os.environ['COMET_GIT_DIRECTORY'] = git_directory
                log_code = True
            else:
                log_code = False

            logger.info("Logging this experiment to comet.ml")

            from tensorboardX import SummaryWriter
            self.comet_writer = SummaryWriter(comet_config={'api_key': api_key,
                                                            'project_name': self.hparams.project_name,
                                                            'log_code': log_code,
                                                            'workspace': self.hparams.comet_workspace,
                                                            'disabled': False})

            self.comet_writer.add_hparams(self.tensorboard_hparams, {})

            self.comet_exp = self.comet_writer._get_comet_logger()._experiment
            self.comet_exp.add_tag(self.hparams.identifier)
            self.comet_exp.set_name(self.exp_name)

        if self.hparams.mlflow:

            from .beam_mlflow import MLflowSummaryWriter
            mlflow_url = self.hparams.mlflow_url
            if mlflow_url is None:
                mlflow_url = f"http://localhost:{beam_service_port('MLFLOW_PORT')}"
            mlflow_exp_name = '-'.join(self.experiment_dir.parts[-4:])
            self.mlflow_writer = MLflowSummaryWriter(mlflow_exp_name, self.tensorboard_hparams, mlflow_url)

        if networks is not None:
            if self.tensorboard_writer is not None:
                for k, net in networks.items():
                    self.tensorboard_writer.add_graph(net, inputs[k])
            if self.comet_exp is not None:
                self.comet_exp.set_model_graph(str(networks))

    def save_model_results(self, reporter, algorithm, iteration, visualize_results=None,
                           store_results=None, store_networks=None, print_results=None,
                           visualize_weights=None, argv=None):

        '''

        responsible for 4 actions:
        1. print results to stdout
        2. visualize results via tensorboard
        3. store results to pandas pickle objects
        4. save networks and optimizers

        logscale is active only on integer epochs in logscale (see x-axis in plt.semilogx)

        :param results:
        :param algorithm:
        :param visualize_results: takes yes|no|logscale.
        :param store_results: takes yes|no|logscale.
        :param store_networks: takes yes|no|logscale.
        :param print_results: whether to print the results to stdout when saving results to tensorboard.
        :return:
        '''

        visualize_weights = self.hparams.get('visualize_weights', preferred=visualize_weights, default=False)
        visualize_results = self.hparams.get('visualize_results', preferred=visualize_results, default='yes')
        store_results = self.hparams.get('store_results', preferred=store_results, default='logscale')
        store_networks = self.hparams.get('store_networks', preferred=store_networks, default='logscale')
        print_results = self.hparams.get('print_results', preferred=print_results, default=True)

        epoch = algorithm.epoch
        if not self.rank:

            if print_results:
                reporter.print_metadata()

            decade = int(np.log10(epoch) + 1)
            logscale = not (epoch - 1) % (self.hparams.visualize_results_log_base ** (decade - 1))

            if ((iteration+1 == algorithm.n_epochs and visualize_results != 'never') or
                    store_results == 'yes' or (store_results == 'logscale' and logscale) or
                (store_results == 'best' and algorithm.best_state)):
                reporter.write_to_path(self.results_dir.joinpath(f'results_{epoch:06d}'))

            alg = algorithm if visualize_weights else None

            if ((iteration+1 == algorithm.n_epochs and visualize_results != 'never') or visualize_results == 'yes' or
                    (visualize_results == 'logscale' and logscale) or
                    (visualize_results == 'best' and algorithm.best_state)):
                self.log_data(reporter, epoch, print_log=print_results, alg=alg, argv=argv)

            if (any([v in store_networks for v in ['yes', 'last', 'logscale']]) or
                (iteration+1 == algorithm.n_epochs and store_networks == 'final') or
                (store_networks == 'all_bests' and algorithm.best_state)):
                checkpoint_file = self.checkpoints_dir.joinpath(f'checkpoint_{epoch:06d}')
                algorithm.save_checkpoint(checkpoint_file)

            if algorithm.best_state and store_networks != 'never':
                checkpoint_file = self.checkpoints_dir.joinpath(f'checkpoint_best')
                algorithm.save_checkpoint(checkpoint_file)

            if 'last' in store_networks or ('logscale' in store_networks and not logscale):
                try:
                    self.checkpoints_dir.joinpath(f'checkpoint_{epoch - 1:06d}').unlink()
                except OSError:
                    pass

    def log_data(self, reporter, n, print_log=True, alg=None, argv=None):

        if print_log:
            reporter.print_stats()

        if self.tensorboard_writer is not None:
            logger.info(f"Tensorboard results are stored to: {self.experiment_dir}")
            reporter.write_to_tensorboard(self.tensorboard_writer, hparams=self.tensorboard_hparams)
        if self.comet_writer is not None:
            logger.info(f"Comet results are stored to: {self.comet_exp.get_key()}")
            reporter.write_to_tensorboard(self.comet_writer, hparams=self.tensorboard_hparams)
        if self.mlflow_writer is not None:
            logger.info(f"MLFlow results are stored to: {self.mlflow_writer.url}")
            reporter.write_to_tensorboard(self.mlflow_writer, hparams=self.tensorboard_hparams)

        def write_network(writer, net, name, n):

            if writer is not None:
                writer.add_histogram("weight_%s/%s" % (net, name), as_numpy(param), n,
                                          bins='tensorflow')
                writer.add_histogram("grad_%s/%s" % (net, name), as_numpy(param.grad), n,
                                          bins='tensorflow')
                if hasattr(param, 'intermediate'):
                    writer.add_histogram("iterm_%s/%s" % (net, name), as_numpy(param.intermediate),
                                              n,
                                              bins='tensorflow')

        if alg is not None:
            networks = alg.networks
            for net in networks:
                for name, param in networks[net].named_parameters():
                    try:
                        write_network(self.tensorboard_writer, net, name, n)
                    except:
                        pass
                    try:
                        write_network(self.comet_writer, net, name, n)
                    except:
                        pass

    @staticmethod
    def _tensorboard(port=None, get_port_from_beam_port_range=True, base_dir=None, log_dirs=None, hparams=False):

        port = find_port(port=port, get_port_from_beam_port_range=get_port_from_beam_port_range)
        if port is None:
            return

        logger.info(f"Opening a tensorboard server on port: {port}")

        if hparams:
            command_argument = f"--bind_all --logdir {base_dir} --port {port}"
        else:
            command_argument = f"--bind_all --logdir_spec={log_dirs} --port {port}"
        from tensorboard.notebook import start as start_tensorboard
        start_tensorboard(command_argument)

    def normalize_experiment_path(self, path, level=0):

        normal_path = [self.hparams.logs_path, self.hparams.project_name,
                       self.hparams.algorithm, self.hparams.identifier]
        pd = path_depth(self.hparams.logs_path)

        return os.path.join(*normal_path[:len(normal_path)-pd-level], path)

    @staticmethod
    def open_tensorboard(root='', project=None, algorithm=None,
                         identifier=None, experiment=None, hparams=False, port=None,
                         get_port_from_beam_port_range=True):
        depth = 4
        filters = {'project': None, 'algorithm': None, 'identifier': None, 'experiment':None}
        if project is not None:
            path_type = check_type(project)
            if path_type.minor == 'list':
                filters['project'] = project
            else:
                filters['project'] = [project]
                path = os.path.join(root, project)
                if os.path.isdir(path):
                    root = path
                    depth = 3

        if algorithm is not None:
            path_type = check_type(algorithm)
            if path_type.minor == 'list':
                filters['algorithm'] = algorithm
            else:
                filters['algorithm'] = [algorithm]
                path = os.path.join(root, algorithm)
                if os.path.isdir(path):
                    root = path
                    depth = 2

        if identifier is not None:
            path_type = check_type(identifier)
            if path_type.minor == 'list':
                filters['identifier'] = identifier
            else:
                filters['identifier'] = [identifier]
                path = os.path.join(root, identifier)
                if os.path.isdir(path):
                    root = path
                    depth = 1
        if experiment is not None:
            path_type = check_type(experiment)
            if path_type.minor == 'list':
                filters['experiment'] = experiment
            else:
                filters['experiment'] = [experiment]
                path = os.path.join(root, experiment)
                if os.path.isdir(path):
                    root = path
                    depth = 0

        experiments = [d[0] for d in list(os.walk(root)) if (path_depth(d[0]) - path_depth(root)) == depth]
        experiments = [os.path.normpath(e) for e in experiments]

        if filters['project'] is not None:
            experiments = list(filter(lambda x: x.split(os.sep)[-4] in filters['project'], experiments))
        if filters['algorithm'] is not None:
            experiments = list(filter(lambda x: x.split(os.sep)[-3] in filters['algorithm'], experiments))
        if filters['identifier'] is not None:
            experiments = list(filter(lambda x: x.split(os.sep)[-2] in filters['identifier'], experiments))
        if filters['experiment'] is not None:
            experiments = list(filter(lambda x: x.split(os.sep)[-1] in filters['experiment'], experiments))

        names = ['/'.join(e.split(os.sep)[-3:]) for e in experiments]
        names = [f"{n}/{gen_hparams_string(e)}" for n, e in zip(names, experiments)]

        experiments = [os.path.join(e, 'tensorboard', 'logs') for e in experiments]
        log_dirs = ','.join([f"{n}:{e}" for n, e in zip(names, experiments)])

        Experiment._tensorboard(port=port, get_port_from_beam_port_range=get_port_from_beam_port_range,
                                base_dir=root, log_dirs=log_dirs, hparams=hparams)

    def tensorboard(self, port=None, add_all_of_same_identifier=False, add_all_of_same_algorithm=False,
                          add_all_of_same_project=False, more_experiments=None, more_identifiers=None,
                          more_algorithms=None, get_port_from_beam_port_range=True, hparams=False):

        suffix = 'hparams' if hparams else 'logs'

        if add_all_of_same_project:
            base_dir = os.path.join(self.hparams.logs_path, self.hparams.project_name)
            depth = 3
        elif add_all_of_same_algorithm:
            base_dir = os.path.join(self.hparams.logs_path, self.hparams.project_name, self.hparams.algorithm)
            depth = 2
        elif add_all_of_same_identifier:
            base_dir = os.path.join(self.hparams.logs_path, self.hparams.project_name, self.hparams.algorithm, self.hparams.identifier)
            depth = 1
        else:
            base_dir = self.experiment_dir
            depth = 0

        base_dir = str(base_dir)
        experiments = [d[0] for d in list(os.walk(base_dir)) if (path_depth(d[0]) - path_depth(base_dir)) == depth]

        if more_experiments is not None:
            if hparams:
                logger.error("hparams visualization does not support adding additional experiments")
            if type(more_experiments) is str:
                more_experiments = [more_experiments]
                experiments = experiments + [self.normalize_experiment_path(e, level=0) for e in more_experiments]

        if more_identifiers is not None:
            if hparams:
                logger.error("hparams visualization does not support adding additional experiments")
            if type(more_identifiers) is str:
                more_identifiers = [more_identifiers]
                depth = 1
                for identifier in more_identifiers:
                    identifier = self.normalize_experiment_path(identifier, level=depth)
                    experiments = experiments + [d[0] for d in list(os.walk(identifier)) if (path_depth(d[0]) - path_depth(identifier)) == depth]

        if more_algorithms is not None:
            if hparams:
                logger.error("hparams visualization does not support adding additional experiments")
            if type(more_algorithms) is str:
                more_algorithms = [more_algorithms]
                depth = 2
                for algorithm in more_algorithms:
                    algorithm = self.normalize_experiment_path(algorithm, level=depth)
                    experiments = experiments + [d[0] for d in list(os.walk(algorithm)) if (path_depth(d[0]) - path_depth(algorithm)) == depth]

        experiments = [os.path.normpath(e) for e in experiments]
        names = ['/'.join(e.split(os.sep)[-3:]) for e in experiments]
        names = [f"{n}/{gen_hparams_string(e)}" for n, e in zip(names, experiments)]

        experiments = [os.path.join(e, 'tensorboard', suffix) for e in experiments]
        log_dirs = ','.join([f"{n}:{e}" for n, e in zip(names, experiments)])

        self._tensorboard(port=port, get_port_from_beam_port_range=get_port_from_beam_port_range,
                          base_dir=base_dir, log_dirs=log_dirs, hparams=hparams)

    def fit(self, alg=None, dataset=None, algorithm_generator=None, return_results=False, reload_results=False,
            tensorboard_arguments=None, alg_args=None, alg_kwargs=None, dataset_args=None,
            dataset_kwargs=None, **kwargs):

        if not self.load_model and not self.logs_path_is_built:
            self.build_experiment_dir()

        if algorithm_generator is None:
            algorithm_generator = beam_algorithm_generator

        try:

            if dataset is not None:
                kwargs['dataset'] = dataset
            if alg_args is not None:
                kwargs['alg_args'] = alg_args
            if alg_kwargs is not None:
                kwargs['alg_kwargs'] = alg_kwargs
            if dataset_args is not None:
                kwargs['dataset_args'] = dataset_args
            if dataset_kwargs is not None:
                kwargs['dataset_kwargs'] = dataset_kwargs
            if tensorboard_arguments is not None:
                kwargs['tensorboard_arguments'] = tensorboard_arguments

            if self.hparams.get('federated_runner'):
                res = self.federated_training(alg=alg, algorithm_generator=algorithm_generator, **kwargs)
            else:
                res = self.run(default_runner, *(algorithm_generator, self, alg), **kwargs)

        except KeyboardInterrupt:

            res = None
            logger.warning(f"KeyboardInterrupt: Training was interrupted, reloads last checkpoint")

        # take care of what is done after training ends
        if res is None or (self.world_size > 1 and not self.hparams.get('federated_runner')):
            alg = algorithm_generator(self, alg, **kwargs)
            results = None
            self.reload_checkpoint(alg)

            if reload_results:
                results = {}
                for subset in alg.results_dir.iterdir():
                    res = list(subset.iterdir())
                    res = pd.DataFrame({'name': res, 'index': [int(c.name.split('_')[-1]) for c in res]})
                    res = res.sort_values('index')

                    res = res.iloc['name']
                    path = alg.results_dir.joinpath(subset, res)
                    results[subset] = path

                if reload_results:
                    results = {subset: path.read() for subset, path in results.items()}

        else:
            alg, results = res

        if return_results:
            return alg, results
        else:
            return alg

    def federated_training(self, alg, algorithm_generator=None, dataset=None, alg_args=None, alg_kwargs=None,
                           dataset_args=None, dataset_kwargs=None,
                           ray_kwargs=None, remote_kwargs=None, **kwargs):

        from ..federated import federated_executor, worker_executor

        if algorithm_generator is None:
            algorithm_generator = self.algorithm_generator

        workers = federated_executor(func=worker_executor, world_size=self.world_size,
                                     framework=self.distributed_training_framework,
                                     distributed_backend=self.hparams.get('distributed_backend'),
                                     host=self.hparams.get('mp_ip'), port=self.hparams.get('mp_port'),
                                     kv_store=self.hparams.get('kv_store'),
                                     kv_store_path=self.hparams.get('kv_store_path'),
                                     kv_store_timeout=self.hparams.get('kv_store_timeout'),
                                     kv_store_port=self.hparams.get('kv_store_port'), ray_address=None,
                                     ray_kwargs=ray_kwargs,
                                     num_gpus=self.hparams.get('n_gpus_per_worker'),
                                     num_cpus=self.hparams.get('n_cpus_per_worker'), remote_kwargs=remote_kwargs, **kwargs)

        if self.world_size > 1:
            logger.info(f'Initializing {self.world_size} parallel workers')
        else:
            logger.info(f'Single worker mode')

        results = []
        for i, we in enumerate(workers):
            logger.info(f"Starting worker {i+1}/{len(workers)} on host: {we.hostname.value} "
                        f"with gpus: {we.physical_devices.value}")
            results.append(we(self, alg, algorithm_generator, dataset=dataset, alg_args=alg_args, alg_kwargs=alg_kwargs,
                              dataset_args=dataset_args, dataset_kwargs=dataset_kwargs))

        return results[0].wait()

    @lazy_property
    def distributed_training_framework(self):
        if self.training_framework in ['accelerate', 'deepspeed']:
            return 'deepspeed'
        return 'ddp'

    @lazy_property
    def default_hparams(self):
        return UniversalConfig(return_defaults=True)

    def build_experiment_dir(self):

        logger.cleanup()
        logger.add_file_handlers(self.experiment_dir.joinpath('experiment.log'))
        print_beam_hyperparameters(self.hparams, default_params=UniversalConfig(return_defaults=True),
                                   debug_only=not self.print_hyperparameters)

        self.experiment_dir.mkdir()
        self.hparams.to_path(self.experiment_dir.joinpath('hparams.pkl'))

        if not self.hparams.override:
            logger.info(f"Creating new experiment")

        else:
            logger.warning("Deleting old experiment")

            self.experiment_dir.rmtree()
            self.exp_name = "%04d_%s" % (self.exp_num, self.exptime)
            self.experiment_dir = self.experiment_dir.parent.joinpath(self.exp_name)

            # set dirs
            self.tensorboard_dir = self.experiment_dir.joinpath('tensorboard')
            self.checkpoints_dir = self.experiment_dir.joinpath('checkpoints')
            self.results_dir = self.experiment_dir.joinpath('results')
            self.code_dir = self.experiment_dir.joinpath('code')

        logger.info(f"Experiment directory is: {self.experiment_dir}")

        self.tensorboard_dir.joinpath('logs').mkdir()
        self.tensorboard_dir.joinpath('hparams').mkdir()
        self.checkpoints_dir.mkdir()

        # make log dirs
        self.results_dir.mkdir()

        # copy code to dir
        if is_notebook():
            code_root_path = os.getcwd()
        else:
            code_root_path = sys.argv[0]

        self.source_dir = os.path.dirname(os.path.realpath(code_root_path))
        if self.hparams.copy_code:
            self.code_dir.copy(beam_path(self.source_dir), include=['.py', '.md', '.ipynb'])

        self.experiment_dir.joinpath('args.pkl').write(self.vars_args)
        self.logs_path_is_built = True

    def run(self, job, *args, **kwargs):

        arguments = (job, self, *args)

        def _run_mpi(demo_fn, world_size):
            from mpi4py import MPI
            comm = MPI.COMM_WORLD
            rank = comm.Get_rank()
            assert rank == 0, "The main process should be rank 0."

            spawned_comm = []
            for i in range(1, world_size):
                # This does not work as it requires the same executable to be run (not a function)
                spawned_comm.append(MPI.COMM_SELF.Spawn(demo_fn, args=(rank, world_size, kwargs, *arguments),
                                                        maxprocs=world_size))

            # Receive results from child processes
            res = [demo_fn(0, world_size, kwargs, *arguments)]
            for i in range(world_size-1):
                result = spawned_comm[i].recv(source=MPI.ANY_SOURCE)
                res.append(result)
                spawned_comm[i].Disconnect()

            return res

        def _run(demo_fn, world_size):

            ctx = mp.get_context(self.hparams.mp_context)
            results_queue = ctx.Queue()
            for rank in range(world_size):
                ctx.Process(target=demo_fn, args=(rank, world_size, results_queue, *arguments),
                            kwargs=kwargs).start()

            res = []
            for rank in range(world_size):
                res.append(results_queue.get())

            return res

        if self.world_size > 1:
            logger.info(f'Initializing {self.world_size} parallel workers')
            if self.distributed_training_framework == 'ddp':
                logger.warning(f"Caution: Sometimes DDP experiments can fail due to a bad configuration. "
                               f"Specifically, if in_place error set --no-broadcast-buffer flag and for subgraph issues"
                               f"set --find-unused-parameters")

            if self.hparams.mp_port is None:
                self.hparams.set('mp_port', find_free_port())
            elif not check_if_port_is_available(self.hparams.mp_port):
                logger.warning(f"Port {self.hparams.mp_port} is not available. Using random port")
                self.hparams.set('mp_port', find_free_port())

            logger.info(f'Multiprocessing port is: {self.hparams.mp_port}')

            if self.hparams.get('distributed_backend') == 'mpi':
                return _run_mpi(run_worker, self.world_size)
            else:
                return _run(run_worker, self.world_size)
        else:
            logger.info(f'Single worker mode')
            return run_worker(0, 1, None, *arguments, **kwargs)

    def writer_cleanup(self):

        if self.tensorboard_writer is not None:
            self.tensorboard_writer.close()
            self.tensorboard_writer = None

        if self.comet_writer is not None:
            self.comet_writer.close()
            self.comet_writer = None

        if self.mlflow_writer is not None:
            self.mlflow_writer.close()
            self.mlflow_writer = None



import json
from collections import namedtuple
from functools import partial
from typing import Optional, Any, Dict, List, Mapping
from uuid import uuid4 as uuid
import pandas as pd
from pydantic import BaseModel

from ..logger import beam_logger as logger
from .response import LLMResponse
from .utils import estimate_tokens, split_to_tokens
# from ..core.processor import Processor
from ..utils import (parse_text_to_protocol, get_edit_ratio, lazy_property, retry, BeamDict, NullClass,
                     pretty_print_dict)
from ..path import BeamURL, BeamResource

LLM = BaseModel
CallbackManagerForLLMRun = NullClass
# import pydantic and check if its version is less than 2.0.0
import pydantic
if pydantic.__version__ < '2.0.0':
    try:
        from langchain.llms.base import LLM
        from langchain.callbacks.manager import CallbackManagerForLLMRun
    except ImportError:
        logger.warning("langchain not found, using pydantic only as the LLM base")
else:
    logger.warning("pydantic version >= 2.0.0 is incompatible with langchain, using pydantic only as the LLM base")

from pydantic import Field, PrivateAttr
from .hf_conversation import Conversation
from .utils import get_conversation_template
from .tools import LLMTool


CompletionObject = namedtuple("CompletionObject", "prompt kwargs response")


class ChatCompletionChunk(BeamDict):
    def __repr__(self):
        return pretty_print_dict(self, 'ChatCompletionChunk')


class ChatCompletion(BeamDict):
    def __repr__(self):
        return pretty_print_dict(self, 'ChatCompletion')


class Completion(BeamDict):
    def __repr__(self):
        return pretty_print_dict(self, 'Completion')


class PedanticBeamResource(LLM, BeamResource):
    url: Optional[BeamURL] = Field(None)
    resource_type: Optional[str] = Field(None)
    scheme: Optional[str] = Field(None)

    # url: BeamURL
    # resource_type: str
    # scheme: str

    class Config:
        arbitrary_types_allowed = True

    def __init__(self, *args, resource_type=None, scheme=None, **kwargs):
        super().__init__(*args, resource_type=resource_type, scheme=scheme, **kwargs)


# class BeamLLM(LLM, Processor, metaclass=CombinedMeta):
class BeamLLM(PedanticBeamResource):

    model: Optional[str] = Field(None)
    scheme: Optional[str] = Field(None)
    usage: Dict = Field(default_factory=dict)
    instruction_history: List = Field(default_factory=list)
    _chat_history: Any = PrivateAttr(default=None)
    _url: Any = PrivateAttr(default=None)
    _debug_langchain: Any = PrivateAttr(default=None)
    temperature: float = Field(1.0, ge=0.0, le=1.0)
    top_p: float = Field(1.0, ge=0.0, le=1.0)
    n: int = Field(1, ge=1)
    message_stream: bool = Field(False)
    stop: Optional[str] = Field(None)
    max_tokens: Optional[int] = Field(None)
    max_new_tokens: Optional[int] = Field(None)
    presence_penalty: float = Field(0.0, ge=-2.0, le=2.0)
    frequency_penalty: float = Field(0.0, ge=-2.0, le=2.0)
    logit_bias: Optional[Dict[str, float]] = Field(None)
    parse_retrials: int = Field(3, ge=0)
    ask_retrials: int = Field(1, ge=1)
    sleep: float = Field(1.0, ge=0.0)
    adapter: Optional[str] = Field(None)
    _len_function: Any = PrivateAttr(default=None)

    # To be used with pydantic classes and lazy_property
    _lazy_cache: Any = PrivateAttr(default=None)
    _tokenizer: Any = PrivateAttr(default=None)
    _path_to_tokenizer: Any = PrivateAttr(default=None)

    _assistant_budget: Any = PrivateAttr(default=None)
    _assistant_docstrings: Any = PrivateAttr(default=None)
    _conv: Any = PrivateAttr(default=None)
    _tools: Any = PrivateAttr(default=None)
    _init_is_done: Any = PrivateAttr(default=None)

    def __init__(self, *args, temperature=.1, top_p=1, n=1, stream=False, stop=None, max_tokens=None, presence_penalty=0,
                 frequency_penalty=0.0, logit_bias=None, scheme='unknown', model=None, max_new_tokens=None, ask_retrials=1,
                 debug_langchain=False, len_function=None, tokenizer=None, path_to_tokenizer=None, parse_retrials=3, sleep=1,
                 adapter=None, tools=None, **kwargs):
        super().__init__(resource_type='llm', scheme=scheme, **kwargs)

        if temperature is not None:
            temperature = float(temperature)
        self.temperature = temperature
        if top_p is not None:
            top_p = float(top_p)
        self.top_p = top_p
        if n is not None:
            n = int(n)
        self.n = n
        self.message_stream = bool(stream)
        self.stop = stop
        if max_tokens is not None:
            max_tokens = int(max_tokens)
        self.max_tokens = max_tokens
        if max_new_tokens is not None:
            max_new_tokens = int(max_new_tokens)
        self.max_new_tokens = max_new_tokens
        if presence_penalty is not None:
            presence_penalty = float(presence_penalty)
        self.presence_penalty = presence_penalty
        if frequency_penalty is not None:
            frequency_penalty = float(frequency_penalty)
        self.frequency_penalty = frequency_penalty

        self.parse_retrials = int(parse_retrials)
        self.ask_retrials = int(ask_retrials)
        self.sleep = float(sleep)
        self.logit_bias = logit_bias
        self.scheme = scheme
        self._url = None
        self._assistant_budget = None
        self._assistant_docstrings = None
        self._debug_langchain = debug_langchain
        self.instruction_history = []

        self.model = model
        if self.model is None:
            self.model = 'unknown'

        if adapter is None:
            adapter = self.model
        self.adapter = adapter

        self._conv = get_conversation_template(self.adapter)

        self.usage = {"prompt_tokens": 0,
                      "completion_tokens": 0,
                      "total_tokens": 0}

        self._chat_history = None

        self._path_to_tokenizer = path_to_tokenizer
        self._tokenizer = tokenizer
        self._len_function = len_function

        self.reset_chat()
        self._lazy_cache = {}
        self._tools = tools

    def add_tool(self, name=None, tool_type='function', func=None, description=None, tool=None, **kwargs):
        if self._tools is None:
            self._tools = {}
        if name in self._tools:
            return
        if type(name) is LLMTool:
            tool = name
        if tool is None:
            tool = LLMTool(name=name, tool_type=tool_type, description=description, func=func, **kwargs)
        self._tools[name] = tool

    def remove_tool(self, name):
        if self._tools is None:
            return
        if name in self._tools:
            self._tools.pop(name)
        if len(self._tools) == 0:
            self._tools = None

    def judge(self, prompt, response, **kwargs):
        # boolean judgement of response by LLM for the prompt

        judge_prompt = (f"Evaluate the following response by an LLM for accuracy and relevance to the given prompt.\n"
                        f"Original Prompt: {prompt}\n"
                        f"LLM Response: {response}\n"
                        f"Based on the factual correctness, relevance to the original prompt, and logical consistency, "
                        f"is the provided response accurate? "
                        f"Respond with 'True' for accurate or 'False' for inaccurate.")

        res = self.ask(judge_prompt, **kwargs)

        return res.bool

    @property
    def tools(self):
        return self._tools

    @property
    def tool_message(self):
        if self.tools is None:
            return ""

        all_tools_message = ""
        for tool in self.tools:
            all_tools_message = f"{all_tools_message}\n\n{tool.name}:\n{tool.dict()}\n"

        message = (f"\nBelow is a list of available tools:\n"
                   f"{all_tools_message}"
                   f"To activate a tool use the following pattern with a JSON formatted argument:\n"
                   f"[TOOL][<tool name>]{{\"args\": [<list of args>], "
                   f"\"kwargs\": {{<a dictionary of kwargs>}}}}[/TOOL]\n\n")
        return message

    @property
    def stop_sequence(self):
        return self._conv.stop_str

    @property
    def sep(self):
        return self._conv.sep

    @property
    def sep2(self):
        return self._conv.sep2

    @property
    def roles(self):
        return self._conv.roles

    @lazy_property
    def tokenizer(self):
        if self._tokenizer is not None:
            tokenizer = self._tokenizer
        elif self._path_to_tokenizer is not None:
            from transformers import PreTrainedTokenizerFast
            tokenizer = PreTrainedTokenizerFast(tokenizer_file=self._path_to_tokenizer)
        else:
            try:
                import tiktoken
                enc = tiktoken.encoding_for_model(self.model).encode
            except ImportError:
                enc = split_to_tokens
            except KeyError:
                try:
                    enc = tiktoken.encoding_for_model("gpt-4").encode
                except:
                    enc = split_to_tokens

            def tokenizer(text):
                return {"input_ids": enc(text)}

        return tokenizer

    @property
    def url(self):

        if self._url is None:
            self._url = BeamURL(scheme=self.scheme, path=self.model)

        return str(self._url)

    @property
    def conversation(self):
        return self._chat_history

    def extract_choices(self, response):
        raise NotImplementedError

    def reset_chat(self):
        self._chat_history = Conversation()

    @property
    def chat_history(self):
        ch = list(self._chat_history.iter_texts())
        return [{'role': 'user' if m[0] else 'assistant', 'content': m[1]} for m in ch]

    def add_to_chat(self, text, is_user=True):
        if is_user:
            self._chat_history.add_user_input(text)
        else:
            self._chat_history.append_response(text)
            self._chat_history.mark_processed()

    @property
    def _llm_type(self) -> str:
        return "beam_llm"

    def _call(self, prompt: str, stop: Optional[List[str]] = None,
              run_manager: Optional[CallbackManagerForLLMRun] = None, **kwargs) -> str:

        res = self.ask(prompt, stop=stop)
        if self._debug_langchain:
            logger.info(f"Prompt: {prompt}")
            logger.info(f"Response: {res.text}")

        return res.text

    @property
    def _identifying_params(self) -> Mapping[str, Any]:
        """Get the identifying parameters."""
        return {"is_chat": self.is_chat,
                'usage': self.usage}

    def len_function(self, prompt):

        if self.tokenizer is not None:
            return len(self.tokenizer(prompt)['input_ids'])

        if self._len_function is None:
            return estimate_tokens(prompt)

        return self._len_function(prompt)

    @property
    def is_chat(self):
        raise NotImplementedError

    @property
    def is_completions(self):
        return not self.is_chat

    def _chat_completion(self, **kwargs):
        raise NotImplementedError

    def add_tool_message_to_chat(self, messages=None):

        if self.tools is None:
            return messages

        if messages is None:
            messages = []

        system_found = False
        for m in messages:
            if m['role'] == 'system':
                m['content'] = f"{m['content']}\n\n{self.tool_message}"
                system_found = True
                break
        if not system_found:
            messages.insert(0, {'role': 'system', 'content': self.tool_message})

        return messages

    def chat_completion(self, stream=False, parse_retrials=3, sleep=1, ask_retrials=1, prompt_type='chat_completion',
                        add_tools_message=False, **kwargs):

        chat_completion_function = self._chat_completion
        if ask_retrials > 1:
            chat_completion_function = retry(func=chat_completion_function, retrials=ask_retrials,
                                        sleep=sleep, logger=logger,
                                        name=f"LLM chat-completion with model: {self.model}")

        tool_names = []
        if self.scheme != 'openai' and add_tools_message and 'tools' in kwargs and kwargs['tools'] is not None:

            # add tools
            tools = kwargs.pop('tools')

            for t in tools:
                tool_type = t['type']
                name = t[tool_type]['name']
                description = t[tool_type]['description']
                parameters = t[tool_type]['parameters']
                tool_names.append(name)
                self.add_tool(name=name, tool_type=tool_type, func=None, description=description, tool=None,
                              **parameters)

            # add tools message
            self.add_tool_message_to_chat(kwargs.pop('messages', None))

        kwargs['stream'] = stream
        completion_object = chat_completion_function(**kwargs)
        response = LLMResponse(completion_object.response, self, chat=True, stream=stream,
                               parse_retrials=parse_retrials, sleep=sleep, prompt_type=prompt_type,
                               prompt_kwargs=completion_object.kwargs, prompt=completion_object.prompt)

        if tool_names:
            for tool_name in tool_names:
                self.remove_tool(tool_name)

        return response

    def completion(self, parse_retrials=3, sleep=1, ask_retrials=1, prompt_type='completion', stream=False, **kwargs):

        completion_function = self._completion
        if ask_retrials > 1:
            completion_function = retry(func=completion_function, retrials=ask_retrials,
                                        sleep=sleep, logger=logger,
                                        name=f"LLM completion with model: {self.model}")

        kwargs['stream'] = stream
        completion_object = completion_function(**kwargs)
        response = LLMResponse(completion_object.response, self, chat=False, stream=stream,
                               parse_retrials=parse_retrials, sleep=sleep, prompt_type=prompt_type,
                               prompt_kwargs=completion_object.kwargs, prompt=completion_object.prompt)
        return response

    def _completion(self, **kwargs):
        raise NotImplementedError

    def get_default_params(self, temperature=None, top_p=None, n=None, stream=None, stop=None, max_tokens=None,
                           presence_penalty=None, frequency_penalty=None, logit_bias=None, max_new_tokens=None,
                           parse_retrials=None, ask_retrials=None, sleep=None):

        if temperature is None:
            temperature = self.temperature
        if top_p is None:
            top_p = self.top_p
        if n is None:
            n = self.n
        if stream is None:
            stream = self.message_stream
        if presence_penalty is None:
            presence_penalty = self.presence_penalty
        if frequency_penalty is None:
            frequency_penalty = self.frequency_penalty
        if logit_bias is None:
            logit_bias = self.logit_bias
        if max_new_tokens is None:
            max_new_tokens = self.max_new_tokens
        if max_tokens is None:
            max_tokens = self.max_tokens
        if ask_retrials is None:
            ask_retrials = self.ask_retrials

        if parse_retrials is None:
            parse_retrials = self.parse_retrials
        if sleep is None:
            sleep = self.sleep

        kwargs = {'temperature': temperature,
                  'top_p': top_p,
                  'n': n,
                  'stream': stream,
                  'stop': stop,
                  'max_tokens': max_tokens,
                  'presence_penalty': presence_penalty,
                  'frequency_penalty': frequency_penalty,
                  'parse_retrials': parse_retrials,
                  'ask_retrials': ask_retrials,
                  'sleep': sleep}

        if max_new_tokens is not None:
            kwargs['max_new_tokens'] = max_new_tokens
        if logit_bias is not None:
            kwargs['logit_bias'] = logit_bias

        return kwargs

    def chat(self, message, name=None, system=None, system_name=None, reset_chat=False, temperature=None,
             top_p=None, n=None, stream=None, stop=None, max_tokens=None, presence_penalty=None, frequency_penalty=None,
             logit_bias=None, max_new_tokens=None, parse_retrials=None, sleep=None, ask_retrials=None, add_tools=True,
             prompt_type='chat_completion', **kwargs):

        '''

        :param name:
        :param system:
        :param system_name:
        :param reset_chat:
        :param temperature:
        :param top_p:
        :param n:
        :param stream:
        :param stop:
        :param max_tokens:
        :param presence_penalty:
        :param frequency_penalty:
        :param logit_bias:
        :return:

        Args:
            message:
        '''

        default_params = self.get_default_params(temperature=temperature,
                                                 top_p=top_p, n=n, stream=stream,
                                                 stop=stop, max_tokens=max_tokens,
                                                 presence_penalty=presence_penalty,
                                                 frequency_penalty=frequency_penalty,
                                                 logit_bias=logit_bias, max_new_tokens=max_new_tokens,
                                                 parse_retrials=parse_retrials, sleep=sleep, ask_retrials=ask_retrials)

        if reset_chat:
            self.reset_chat()

        if type(message) is list:
            self.reset_chat()
            messages = message
        else:
            messages = []

            if system is not None:
                system = {'role': 'system', 'content': system}
                if system_name is not None:
                    system['system_name'] = system_name
                messages.append(system)

            messages.extend(self.chat_history)

        self.add_to_chat(message, is_user=True)
        message = {'role': 'user', 'content': message}
        if name is not None:
            message['name'] = name

        messages.append(message)

        if add_tools:
            self.add_tool_message_to_chat(messages)

        response = self.chat_completion(messages=messages, prompt_type=prompt_type, **default_params)

        if stream:
            def gen():

                generated_text = []
                for res in response:
                    if res.text is not None:
                        generated_text.append(res.text)
                    yield res

                generated_text = "".join(generated_text)
                self.update_usage(res.response)
                self.add_to_chat(generated_text, is_user=False)

            return gen()

        else:
            self.update_usage(response.response)
            self.add_to_chat(response.text, is_user=False)
            return response

    def chat_with_assistant(self, message=None, docstrings=None, budget=3, **kwargs):

        if docstrings is not None:
            self._assistant_budget = budget
            self._assistant_docstrings = docstrings
            self.reset_chat()
        else:
            budget = self._assistant_budget
            docstrings = self._assistant_docstrings

        docs = ""
        for i, (k, v) in enumerate(docstrings.items()):
            docs += (f"{i}. Function: {k}\n\n{v}\n\n"
                     f"[SEP]\n")

        system = (f"You are given a list of docstrings of several functions. Each docstring details the function name, "
                  f"its purpose, its arguments and its keyworded arguments.\n"
                  f"Your task is to handle a chat session with a user. The user will ask you to apply some functionality "
                  f"or execute some action. You need to chose the function you see fit out of the list of available functions "
                  f"and you need to fill its arguments according to the user request. If the request of the user is unclear and "
                  f"you are not able to assign a proper function, you can ask for clarification. If some of the arguments which "
                  f"are required in the function are missing from the user request, you should ask he or she to provide them. "
                  f"If you are confident that you can assign a function to the user request and fill all the required arguments, "
                  f"and fill as much as possible keyworded arguments, you need to respond in a single valid JSON object "
                  f"of the form {{\"function\": <function name>, \"args\": [list of arguments], \"kwargs\": {{<dictionary of kwargs>}} }}\n\n"
                  f"[SEP]\n\n"
                  f"{docs}\n")

        if message is None:
            message = "Hello, how can you help me?"

        if len(self.chat_history) < 2 * budget - 2:
            res = self.chat(message, system=system, **kwargs)
        else:

            self.add_to_chat(message, is_user=True)

            instruction = (
                f"You are given a list of docstrings of several functions. Each docstring details the function name, "
                f"its purpose, its arguments and its keyworded arguments.\n"
                f"In addition, you are given a chat session between a user and a chatbot assistant. The user asks the chatbot "
                f"apply some functionality or execute some action. You need to chose the function you see fit out"
                f" of the list of available functions and to fill its arguments and as much as possible "
                f"keyworded arguments according to the user request. You are required to respond in a single "
                f"valid JSON object of the form {{\"function\": <function name>, \"args\": [list of arguments], "
                f"\"kwargs\": {{<dictionary of kwargs>}} }}\n\n"
                f"[SEP]\n\n"
                f"Docstrings:\n\n"
                f"{docs}\n"
                f"Chat session:\n\n"
                f"{self.chat_history}"
                f"[SEP]\n\n"
                f"Response: \"\"\"\n{{text input here}}\n\"\"\"")

            res = self.ask(instruction, system=system, **kwargs)

        return res

    def explain_traceback(self, traceback, n_words=100, **kwargs):
        prompt = (f"Task: explain the following traceback and suggest a correction. Be concise don't use more than"
                  f"{n_words} words and don't use newlines.\n\n"
                    f"[SEP]\n\n"
                    f"{traceback}\n\n"
                    f"[SEP]\n\n"
                    f"Response: \"\"\"\n{{text input here}}\n\"\"\"")

        res = self.ask(prompt, **kwargs)
        return res.text

    def fix_protocol(self, text, protocol='json', **kwargs):

        protocols_text = {'json': 'JSON', 'html': 'HTML', 'xml': 'XML', 'csv': 'CSV', 'yaml': 'YAML', 'toml': 'TOML'}

        prompt = (f"Task: fix the following {protocols_text[protocol]} object. "
                  f"Return a valid {protocols_text[protocol]} object without anything else\n\n"
                    f"[SEP]\n\n"
                    f"{text}\n\n"
                    f"[SEP]\n\n"
                    f"Response: \"\"\"\n{{text input here}}\n\"\"\"")
        res = self.ask(prompt).text

        res = parse_text_to_protocol(res, protocol=protocol)

        return res

    def docstring(self, text, element_type, name=None, docstring_format=None, parent=None, parent_name=None,
                  parent_type=None, children=None, children_type=None, children_name=None, **kwargs):

        if docstring_format is None:
            docstring_format = f"in \"{docstring_format}\" format, "
        else:
            docstring_format = ""

        prompt = f"Task: write a full python docstring {docstring_format}for the following {element_type}\n\n" \
                 f"[SEP]\n\n" \
                 f"{text}\n\n" \
                 f"[SEP]\n\n"

        if parent is not None:
            prompt = f"{prompt}" \
                     f"where its parent {parent_type}: {parent_name}, has the following docstring\n\n" \
                     f"[SEP]\n\n" \
                     f"{parent}\n\n" \
                     f"[SEP]\n\n"

        if children is not None:
            for i, (c, cn, ct) in enumerate(zip(children, children_name, children_type)):
                prompt = f"{prompt}" \
                         f"and its #{i} child: {ct} named {cn}, has the following docstring\n\n" \
                         f"[SEP]\n\n" \
                         f"{c}\n\n" \
                         f"[SEP]\n\n"

        prompt = f"{prompt}" \
                 f"Response: \"\"\"\n{{docstring text here (do not add anything else)}}\n\"\"\""

        if not self.is_completions:
            try:
                res = self.chat(prompt, **kwargs)
            except Exception as e:
                logger.error(f"Error in response: {e}")
                try:
                    logger.warning(f"{name}: switching to gpt-4 model")
                    res = self.chat(prompt, model='gpt-4', **kwargs)
                except:
                    logger.error(f"{name}: error in response")
                    res = None
        else:
            res = self.ask(prompt, **kwargs)

        # res = res.choices[0].text

        return res

    def openai_format(self, res, finish_reason="stop", tokens=None, completion_tokens=0, prompt_tokens=0, total_tokens=0):

        stream = res.stream
        text = self.extract_text(res)

        # an example of openai format:
        # ChatCompletionChunk(id='chatcmpl-8ooipEUzmkwlqMVnUJHZzWQ4uKyKj', choices=[
        #     Choice(delta=ChoiceDelta(content=None, function_call=None, role=None, tool_calls=None),
        #            finish_reason='stop', index=0, logprobs=None)], created=1707122067, model='gpt-4-0613',
        #                     object='chat.completion.chunk', system_fingerprint=None)
        #
        # Completion(id='cmpl-8op1yf5pOL7oEd7EXulOFbODmRSFp', choices=[
        #     CompletionChoice(finish_reason='length', index=0, logprobs=None,
        #                      text='\n\nI am an AI and do not have emotions, but thank you for asking')],
        #            created=1707123254, model='gpt-3.5-turbo-instruct', object='text_completion',
        #            system_fingerprint=None,
        #            usage=CompletionUsage(completion_tokens=16, prompt_tokens=4, total_tokens=20))

        if stream:
            Class = ChatCompletionChunk
            choice = {
                "index": 0,
                "delta": {"content": text,
                          "function_call": None,
                          "role": None,
                          "tool_calls": None},
                "logprobs": None,
                "finish_reason": finish_reason,
            }

        elif res.chat:
            Class = ChatCompletion

            if len(self.tools):

                tool_calls = []
                for t in self.tools:
                    et = t(res)
                    if et is not None and et.success:

                        arguments = {}
                        for i, k in enumerate(et.args):
                            arguments[t.args[i]] = k
                        for k, v in et.kwargs.items():
                            arguments[k] = v

                        tool_res = {
                            "id": f"call_{uuid()}",
                            "type": "function",
                            "function": {
                                "name": t.name,
                                "arguments": json.dumps(arguments)
                            }}
                        tool_calls.append(tool_res)

            if len(tool_calls):
                choice = {
                    "finish_reason": finish_reason,
                    "index": 0,
                    "message": {
                        "content": None,
                        "role": "assistant",
                        "tool_calls": tool_calls
                    }
                }
            else:

                choice = {
                          "finish_reason": finish_reason,
                          "index": 0,
                          "message": {
                            "content": text,
                            "role": "assistant"
                          }
                        }
        else:

            Class = Completion
            choice = {
                "finish_reason": finish_reason,
                "index": 0,
                "logprobs": tokens,
                "text": text
            }

        res = {
                  "choices": [
                    choice
                  ],
                  "created": res.created,
                  "id": res.id,
                  "model": res.model,
                  "object": res.object,
                  "system_fingerprint": None,
                  "usage": {
                    "completion_tokens": completion_tokens,
                    "prompt_tokens": prompt_tokens,
                    "total_tokens": total_tokens
                  }
                }

        return Class(**res)

    def ask(self, question, max_tokens=None, temperature=None, top_p=None, frequency_penalty=None, max_new_tokens=None,
            presence_penalty=None, stop=None, n=None, stream=None, logprobs=None, logit_bias=None, echo=False,
            parse_retrials=None, sleep=None, ask_retrials=None, prompt_type='completion', add_tools=True, **kwargs):
        """

        @param question:
        @param max_tokens:
        @param temperature:
        @param top_p:
        @param frequency_penalty:
        @param max_new_tokens:
        @param presence_penalty:
        @param stop:
        @param n:
        @param stream:
        @param logprobs:
        @param logit_bias:
        @param echo:
        @param parse_retrials:
        @param sleep:
        @param ask_retrials:
        @param prompt_type:
        @param kwargs:
        @return:
        """
        default_params = self.get_default_params(temperature=temperature,
                                                 top_p=top_p, n=n, stream=stream,
                                                 stop=stop, max_tokens=max_tokens,
                                                 presence_penalty=presence_penalty,
                                                 frequency_penalty=frequency_penalty,
                                                 logit_bias=logit_bias,
                                                 max_new_tokens=max_new_tokens, ask_retrials=ask_retrials,
                                                 parse_retrials=parse_retrials, sleep=sleep)

        # if response_format is not None:
        #     question = f"{question}\nReplay with a valid {response_format} format"

        if add_tools:
            self.add_tool_message_to_prompt(question)

        if not self.is_completions:
            kwargs = {**default_params, **kwargs}
            response = self.chat(question, reset_chat=True, prompt_type=f'simulated_{prompt_type}_with_chat', **kwargs)
        else:
            response = self.completion(prompt=question, logprobs=logprobs, echo=echo,
                                       prompt_type=prompt_type, **default_params)

        if stream:
            def gen():
                for res in response:
                    yield res

                if not self.is_completions:
                    # currently openai stream does not support update usage
                    self.update_usage(res.response)
                    self.instruction_history.append({'question': question, 'response': res.response, 'type': 'ask'})

            return gen()

        else:
            self.update_usage(response.response)
            self.instruction_history.append({'question': question, 'response': response, 'type': 'ask'})
            return response

    def reset_instruction_history(self):
        self.instruction_history = []

    def summary(self, text, n_words=100, n_paragraphs=None, **kwargs):
        """
        Summarize a text
        :param text:  text to summarize
        :param n_words: number of words to summarize the text into
        :param n_paragraphs:   number of paragraphs to summarize the text into
        :param kwargs: additional arguments for the ask function
        :return: summary
        """
        if n_paragraphs is None:
            prompt = f"Task: summarize the following text into {n_words} words\nText: {text}\nResponse: \"\"\"\n{{text input here}}\n\"\"\""
        else:
            prompt = f"Task: summarize the following text into {n_paragraphs} paragraphs\nText: {text}\nResponse: \"\"\"\n{{text input here}}\n\"\"\""

        res = self.ask(prompt, **kwargs).text
        return res

    def verify_response(self, res):
        return True

    def extract_text(self, res):
        raise NotImplementedError

    def question(self, text, question, **kwargs):
        """
        Answer a yes-no question
        :param text: text to answer the question from
        :param question: question to answer
        :param kwargs: additional arguments for the ask function
        :return: answer
        """
        prompt = f"Task: answer the following question\nText: {text}\nQuestion: {question}\nResponse: \"\"\"\n{{text input here}}\n\"\"\""

        res = self.ask(prompt, **kwargs).text
        return res

    def yes_or_no(self, question, text=None, **kwargs):
        """
        Answer a yes or no question
        :param text: text to answer the question from
        :param question:  question to answer
        :param kwargs: additional arguments for the ask function
        :return: answer
        """

        if text is None:
            preface = ''
        else:
            preface = f"Text: {text}\n"

        prompt = f"{preface}Task: answer the following question with yes or no\nQuestion: {question}\nResponse: \"\"\"\n{{text input here}}\n\"\"\""

        res = self.ask(prompt, **kwargs).text

        res = res.lower().strip()
        res = res.split(" ")[0]

        i = pd.Series(['no', 'yes']).apply(partial(get_edit_ratio, s2=res)).idxmax()
        return bool(i)

    def quant_analysis(self, text, source=None, **kwargs):
        """
        Perform a quantitative analysis on a text
        :param text: text to perform the analysis on
        :param kwargs: additional arguments for the ask function
        :return: analysis
        """
        prompt = f"Task: here is an economic news article from {source}\nText: {text}\nResponse: \"\"\"\n{{text input here}}\n\"\"\""
        res = self.ask(prompt, **kwargs).text
        return res

    def names_of_people(self, text, **kwargs):
        """
        Extract names of people from a text
        :param text: text to extract names from
        :param kwargs: additional arguments for the ask function
        :return: list of names
        """
        prompt = f"Task: extract names of people from the following text, return in a list of comma separated values\nText: {text}\nResponse: \"\"\"\n{{text input here}}\n\"\"\""

        res = self.ask(prompt, **kwargs).text
        res = res.strip().split(",")

        return res

    def answer_email(self, input_email_thread, responder_from, receiver_to, **kwargs):
        """
        Answer a given email thread as an chosen entity
        :param input_email_thread_test: given email thread to answer to
        :param responder_from: chosen entity name which will answer the last mail from the thread
        :param receiver_to: chosen entity name which will receive the generated mail
        :param kwargs: additional arguments for the prompt
        :return: response mail
        """

        prompt = f"{input_email_thread}\n---generate message---\nFrom: {responder_from}To: {receiver_to}\n\n###\n\n"
        # prompt = f"Text: {text}\nTask: answer the following question with yes or no\nQuestion: {question}\nResponse: \"\"\"\n{{text input here}}\n\"\"\""

        res = self.ask(prompt, **kwargs).text
        return res

    def classify(self, text, classes, **kwargs):
        """
        Classify a text
        :param text: text to classify
        :param classes: list of classes
        :param kwargs: additional arguments for the ask function
        :return: class
        """
        prompt = f"Task: classify the following text into one of the following classes\nText: {text}\nClasses: {classes}\nResponse: \"\"\"\n{{text input here}}\n\"\"\""

        res = self.ask(prompt, **kwargs).text
        res = res.lower().strip()

        i = pd.Series(classes).str.lower().str.strip().apply(partial(get_edit_ratio, s2=res)).idxmax()

        return classes[i]

    def features(self, text, features=None, **kwargs):
        """
        Extract features from a text
        :param text: text to extract features from
        :param kwargs: additional arguments for the ask function
        :return: features
        """

        if features is None:
            features = []

        features = [f.lower().strip() for f in features]

        prompt = f"Task: Out of the following set of terms: {features}\n" \
                 f"list in comma separated values (csv) the terms that describe the following Text:\n" \
                 f"[SEP]\n" \
                 f" {text}\n" \
                 f"[SEP]\n" \
                 f"Important: do not list any other term that did not appear in the aforementioned list.\n" \
                 f"Response: \"\"\"\n{{text input here}}\n\"\"\""

        res = self.ask(prompt, **kwargs).text

        llm_features = res.split(',')
        llm_features = [f.lower().strip() for f in llm_features]
        features = [f for f in llm_features if f in features]

        return features

    def entities(self, text, humans=True, **kwargs):
        """
        Extract entities from a text
        :param humans:  if True, extract people, else extract all entities
        :param text: text to extract entities from
        :param kwargs: additional arguments for the ask function
        :return: entities
        """
        if humans:
            prompt = f"Task: extract people from the following text in a comma separated list\nText: {text}\nResponse: \"\"\"\n{{text input here}}\n\"\"\""
        else:
            prompt = f"Task: extract entities from the following text in a comma separated list\nText: {text}\nResponse: \"\"\"\n{{text input here}}\n\"\"\""

        res = self.ask(prompt, **kwargs).text

        entities = res.split(',')
        entities = [e.lower().strip() for e in entities]

        return entities

    def title(self, text, n_words=None, **kwargs):
        """
        Extract title from a text
        :param text: text to extract title from
        :param kwargs: additional arguments for the ask function
        :return: title
        """
        if n_words is None:
            prompt = f"Task: extract title from the following text\nText: {text}\nResponse: \"\"\"\n{{text input here}}\n\"\"\""
        else:
            prompt = f"Task: extract title from the following text. Restrict the answer to {n_words} words only." \
                     f"\nText: {text}\nResponse: \"\"\"\n{{text input here}}\n\"\"\""

        res = self.ask(prompt, **kwargs).text

        return res

    def similar_keywords(self, text, keywords, **kwargs):
        """
        Find similar keywords to a list of keywords
        :param text: text to find similar keywords from
        :param keywords: list of keywords
        :param kwargs: additional arguments for the ask function
        :return: list of similar keywords
        """

        keywords = [e.lower().strip() for e in keywords]
        prompt = f"Keywords: {keywords}\nTask: find similar keywords in the following text\nText: {text}\n\nResponse: \"\"\"\n{{text input here}}\n\"\"\""

        res = self.ask(prompt, **kwargs).text
        res = res.split(',')
        res = [e.lower().strip() for e in res]

        res = list(set(res) - set(keywords))

        return res

    def is_keyword_found(self, text, keywords, **kwargs):
        """
        check if one or more key words found in given text
        :param text: text to looks for
        :param keywords:  key words list
        :param kwargs: additional arguments for the ask function
        :return: yes if one of the keywords found else no
        """
        prompt = f"Text: {text}\nTask: answer with yes or no if Text contains one of the keywords \nKeywords: {keywords}\nResponse: \"\"\"\n{{text input here}}\n\"\"\""

        res = self.ask(prompt, **kwargs).text
        res = res.lower().strip().replace('"', "")

        i = pd.Series(['no', 'yes']).apply(partial(get_edit_ratio, s2=res)).idxmax()
        return bool(i)

    def get_similar_terms(self, keywords, **kwargs):
        """
        check if one or more key words found in given text
        :param keywords:  key words list
        :param kwargs: additional arguments for the ask function
        :return: similar terms
        """
        prompt = f"keywords: {keywords}\nTask: return all semantic terms for given Keywords \nResponse: \"\"\"\n{{text input here}}\n\"\"\""

        res = self.ask(prompt, **kwargs).text
        res = res.lower().strip()
        return res

    def update_usage(self, response):
        pass

    def stem_message(self, message, n_tokens, skip_special_tokens=True):
        if self.tokenizer is not None and hasattr(self.tokenizer, 'decode'):
            tokens = self.tokenizer(message)['input_ids'][:n_tokens+1]  # +1 for the start token
            return self.tokenizer.decode(tokens, skip_special_tokens=skip_special_tokens)
        else:
            return ''.join(split_to_tokens(message)[:n_tokens])

    def add_tool_message_to_prompt(self, question):
        if self.tools is not None:
            question = f"{question}\n\n{self.tool_message}"
        return question



from contextlib import contextmanager
from uuid import uuid4 as uuid
from .resource import beam_path


@contextmanager
def local_copy(path, tmp_path='/tmp', as_beam_path=False, override=False, disable=None):

    if disable is None:
        disable = path.scheme == 'file'

    if disable:
        if as_beam_path:
            yield path
        else:
            yield str(path)
        return

    path = beam_path(path)
    tmp_dir = beam_path(tmp_path).joinpath(uuid())
    tmp_dir.mkdir(exist_ok=True, parents=True)
    tmp_path = tmp_dir.joinpath(path.name)

    exists = path.exists() and (path.is_file() or len(list(path)) > 0)

    assert not exists or override, f'Path {path} already exists, set override=True to overwrite it.'

    if exists:
        path.copy(tmp_path)

    try:
        yield tmp_path if as_beam_path else str(tmp_path)
    finally:

        tmp_path.copy(path)
        tmp_dir.rmtree()


class FileSystem:
    def __init__(self, path):
        self.path = beam_path(path)

    def exists(self):
        return self.path.exists()

    def is_file(self):
        return self.path.is_file()

    def is_dir(self):
        return self.path.is_dir()

    def joinpath(self, path):
         self.path = self.path.joinpath(path)

    def rmtree(self):
        self.path.rmtree()

    def mkdir(self, exist_ok=True, parents=True):
        self.path.mkdir(exist_ok=exist_ok, parents=parents)

    def copy(self, path):
        path = beam_path(path)
        path.parent.mkdir(exist_ok=True, parents=True)
        self.path.copy(path)

    def open(self, mode='r'):
        return self.path.open(mode=mode)

    def root(self):
        self.path = self.path.root()

    def __str__(self):
        return str(self.path)

    def __repr__(self):
        return repr(self.path)

    def __iter__(self):
        for p in self.path:
            yield p.name

    def iterdir(self):
        return self.__iter__()


from ..config import UniversalConfig, BeamParam


class TabularConfig(UniversalConfig):

    defaults = dict(project_name='deep_tabular', algorithm='TabularNet', n_epochs=100, scheduler='one_cycle',
                    batch_size=512, lr_dense=2e-3, lr_sparse=2e-2, early_stopping_patience=16)

    parameters = [
        BeamParam('emb_dim', int, 128, 'latent embedding dimension', tags='tune'),
        BeamParam('n_transformer_head', int, 4, 'number of transformer heads', tags='tune'),
        BeamParam('n_encoder_layers', int, 4, 'number of encoder layers', tags='tune'),
        BeamParam('n_decoder_layers', int, 4, 'number of decoder layers', tags='tune'),
        BeamParam('transformer_hidden_dim', int, 256, 'transformer hidden dimension', tags='tune'),
        BeamParam('transformer_dropout', float, 0., 'transformer dropout', tags='tune'),
        BeamParam('mask_rate', float, 0.15, 'rate of masked features during training', tags='tune'),
        BeamParam('rule_mask_rate', float, 0., 'rate of masked rules during training', tags='tune'),
        BeamParam('maximal_mask_rate', float, 0.2, 'the maximal mask rate with dynamic masking', tags='tune'),
        BeamParam('minimal_mask_rate', float, 0.1, 'the minimal mask rate with dynamic masking', tags='tune'),
        BeamParam('dynamic_delta', float, 0.005, 'the incremental delta for dynamic masking', tags='tune'),
        BeamParam('n_rules', int, 64, 'number of transformers rules in the decoder', tags='tune'),
        BeamParam('activation', str, 'gelu', 'transformer activation', tags='tune'),
        BeamParam('n_quantiles', int, 10, 'number of quantiles for the quantile embeddings', tags='tune'),
        BeamParam('scaler', str, 'quantile', 'scaler for the preprocessing [robust, quantile]', tags='tune'),
        BeamParam('n_ensembles', int, 32, 'number of ensembles of the model for prediction in inference mode', tags='tune'),
        BeamParam('label_smoothing', float, 0., 'label smoothing for the cross entropy loss', tags='tune'),
        BeamParam('dropout', float, .0, 'Output layer dropout of the model', tags='tune'),
        BeamParam('oh_to_cat', bool, False, 'Try to convert one-hot encoded categorical features to categorical features', tags='tune'),
        BeamParam('dynamic_masking', bool, False, 'Use dynamic masking scheduling', tags='tune'),
        BeamParam('feature_bias', bool, True, 'Add bias to the features', tags='tune'),
        BeamParam('rules_bias', bool, True, 'Add bias to the rules', tags='tune'),
        BeamParam('lin_version', int, 1, 'version of the linear output layer', tags='tune'),

        BeamParam('dataset_name', str, 'covtype',
               'dataset name [year, california_housing, higgs_small, covtype, aloi, adult, epsilon, '
               'microsoft, yahoo, helena, jannis]'),
        BeamParam('catboost', bool, False, 'Train a catboost model on the data'),
        BeamParam('store_data_on_device', bool, True, 'Store the data on the device (GPU/CPU) in advance'),
        BeamParam('rulenet', bool, True, 'Train our RuleNet model on the data'),
        BeamParam('sparse-embedding', bool, True, 'Use sparse embedding for the features embeddings'),
    ]




from .models import (BeamPath, S3Path, S3PAPath, HDFSPath, HDFSPAPath, SFTPPath, CometAsset,
                     RedisPath, SMBPath)
from .core import BeamKey, BeamURL, IOPath, DictPath


beam_key = BeamKey()


def beam_path(path, username=None, hostname=None, port=None, private_key=None, access_key=None, secret_key=None,
              password=None, **kwargs):
    """

    @param port:
    @param hostname:
    @param username:
    @param protocol:
    @param private_key:
    @param secret_key: AWS secret key
    @param access_key: AWS access key
    @param path: URI syntax: [protocol://][username@][hostname][:port][/path/to/file]
    @return: BeamPath object
    """
    if type(path) != str:
        return path

    if ':' not in path:
        return BeamPath(path, scheme='file')
    elif path[1] == ':':  # windows path
        path = path.replace('\\', '/')
        path = path.lstrip('/')
        return BeamPath(path, scheme='nt')

    url = BeamURL.from_string(path)

    if url.hostname is not None:
        hostname = url.hostname

    if url.port is not None:
        port = url.port

    if url.username is not None:
        username = url.username

    if url.password is not None:
        password = url.password

    query = url.query
    for k, v in query.items():
        kwargs[k] = v

    if access_key is None and 'access_key' in kwargs:
        access_key = kwargs.pop('access_key')

    if private_key is None and 'private_key' in kwargs:
        private_key = kwargs.pop('private_key')

    if secret_key is None and 'secret_key' in kwargs:
        secret_key = kwargs.pop('secret_key')

    path = url.path

    if url.protocol is None or (url.protocol == 'file'):
        return BeamPath(path)

    if path == '':
        path = '/'

    username = beam_key('BEAM_USERNAME', username)
    password = beam_key('BEAM_PASSWORD', password)

    if 's3' in url.protocol:

        access_key = beam_key('AWS_ACCESS_KEY_ID', access_key)
        secret_key = beam_key('AWS_SECRET_ACCESS_KEY', secret_key)

        if url.protocol == 's3-pa':
            return S3PAPath(path, hostname=hostname, port=port, access_key=access_key, secret_key=secret_key, **kwargs)
        else:
            return S3Path(path, hostname=hostname, port=port, access_key=access_key, secret_key=secret_key,  **kwargs)

    elif url.protocol == 'hdfs':
        return HDFSPath(path, hostname=hostname, port=port, username=username, password=password, **kwargs)

    elif url.protocol == 'hdfs-pa':
        return HDFSPAPath(path, hostname=hostname, port=port, username=username, password=password, **kwargs)

    elif url.protocol == 'redis':
        return RedisPath(path, hostname=hostname, port=port, username=username, password=password, **kwargs)

    elif url.protocol == 'smb':
        return SMBPath(path, hostname=hostname, port=port, username=username, password=password, **kwargs)

    elif url.protocol == 'comet':

        access_key = beam_key('COMET_API_KEY', access_key)
        return CometAsset(path, access_key=access_key, **kwargs)

    elif url.protocol == 'io':
        return IOPath(path, **kwargs)

    elif url.protocol == 'dict':
        return DictPath(path, **kwargs)

    elif url.protocol == 'gs':
        raise NotImplementedError
    elif url.protocol == 'http':
        raise NotImplementedError
    elif url.protocol == 'https':
        raise NotImplementedError
    elif url.protocol == 'ftp':
        raise NotImplementedError
    elif url.protocol == 'ftps':
        raise NotImplementedError
    elif url.protocol == 'nt':
        path = path.replace('\\', '/')
        return BeamPath(path)
    elif url.protocol == 'sftp':

        private_key = beam_key('SSH_PRIVATE_KEY', private_key)
        return SFTPPath(path, hostname=hostname, username=username, port=port,
                        private_key=private_key, password=password, **kwargs)
    else:
        raise NotImplementedError


def in_memory_storage(mode=None, data=None):
    if mode == 'file':
        return beam_path('io:///', data=data)
    return beam_path('dict:///', data=data)



import io
import json
from threading import Thread
from uuid import uuid4 as uuid
import websocket
from ..logger import beam_logger as logger

from ..path import normalize_host
from ..serve.http_client import HTTPClient


class AsyncClient(HTTPClient):

    def __init__(self, hostname=None, ws_port=None, ws_tls=False, postrun=None, enable_websocket=True,  **kwargs):
        super().__init__(hostname=hostname, **kwargs)

        metadata = self.info.get('metadata', {})
        if ws_port is None and 'ws_port' in metadata:
            ws_port = metadata['ws_port']
            self.ws_application = metadata['ws_application']
        else:
            self.ws_application = 'ws' if not ws_tls else 'wss'

        self.ws_host = normalize_host(hostname, ws_port)
        if postrun is None:
            self.postrun_callback = self.postrun
        else:
            self.postrun_callback = postrun

        self.client_id = f"ws-client-{uuid()}"

        if enable_websocket:
            self.init_websocket()

        self.ws = None
        self.wst = None

    def set_callback(self, callback):
        self.postrun_callback = callback

    def post(self, path, *args, postrun_args=None, postrun_kwargs=None, **kwargs):

        io_args = io.BytesIO()
        self.dump_function(args, io_args)
        io_args.seek(0)

        io_kwargs = io.BytesIO()
        self.dump_function(kwargs, io_kwargs)
        io_kwargs.seek(0)

        io_postrun_args = io.BytesIO()
        self.dump_function(postrun_args, io_postrun_args)
        io_postrun_args.seek(0)

        io_postrun_kwargs = io.BytesIO()
        self.dump_function(postrun_kwargs, io_postrun_kwargs)
        io_postrun_kwargs.seek(0)

        io_ws_client_id = io.BytesIO()
        self.dump_function(self.client_id, io_ws_client_id)
        io_ws_client_id.seek(0)

        response = self._post(path, io_args, io_kwargs, postrun_args=io_postrun_args,
                              postrun_kwargs=io_postrun_kwargs, ws_client_id=io_ws_client_id)

        return response

    def init_websocket(self):
        self.ws = websocket.WebSocketApp(f"{self.ws_application}://{self.ws_host}/",
                                         on_open=self.on_open,
                                         on_message=self.on_message,
                                         on_error=self.on_error,
                                         on_close=self.on_close)
        # Run the WebSocket client in a separate thread
        self.wst = Thread(target=self.ws.run_forever)
        self.wst.start()

    def postrun(self, result):
        pass

    def poll(self, task_id, timeout=None):
        params = {'task_id': task_id}
        if timeout is not None:
            params['timeout'] = timeout
        return self.get('/poll/beam', params=params)

    def on_message(self, ws, message):
        data = json.loads(message)

        # Extract task_id and state from the message
        task_id = data.get('task_id')
        state = data.get('state')
        metadata = data.get('metadata', None)

        if state == 'SUCCESS':
            result = self.poll(task_id)
            self.postrun_callback(result)
        else:
            logger.error(f"Task {task_id} failed with state {state} and metadata: {metadata}")

    def on_error(self, ws, error):
        logger.error(f"WebSocket error: {error}")

    def on_close(self, ws, close_status_code, close_msg):
        logger.warning(f"WebSocket closed with status code {close_status_code} and message: {close_msg}")
        self.init_websocket()

    def on_open(self, ws):
        logger.info(f"Opening websocket at {self.ws_application}://{self.ws_host}/: client_id: {self.client_id}")
        ws.send(self.client_id)



from typing import Union


class K8SUnits:
    # manage transformation from numbers to k8s units (Gi, m, etc.)
    def __init__(self, value):
        if isinstance(value, int):
            self.value = value
        elif isinstance(value, K8SUnits):
            self.value = value.value
        elif isinstance(value, str):
            self.value = self.parse_str(value)

    def parse_str(self, value: str):
        if 'Gi' in value:
            return int(value.replace('Gi', '')) * 1000 ** 3
        elif 'm' in value:
            return int(value.replace('m', '')) * 1000 ** 2
        else:
            return int(value)

    @property
    def as_str(self):
        if self.value // 1000 ** 3 == self.value / 1000 ** 3 and self.value >= 1000 ** 3:
            return f"{self.value / 1000 ** 3}Gi"
        elif self.value // 1000 ** 2 == self.value / 1000 ** 2 and self.value >= 1000 ** 2:
            return f"{self.value / 1000 ** 2}m"
        elif self.value is None:
            return None
        else:
            return str(self.value)

    @property
    def as_number(self):
        return self.value

    def __str__(self):
        return self.as_str

    def __int__(self):
        return self.value



import copy
import time

from .params import HPOConfig
from ..utils import find_port, check_type, is_notebook, beam_device
from ..config import print_beam_hyperparameters
from ..logger import beam_logger as logger
from ..path import beam_path, BeamPath
from ..core import Processor
from ..algorithm import NeuralAlgorithm
from functools import partial
from ..experiment import beam_algorithm_generator

from .._version import __version__


class BeamHPO(Processor):

    def __init__(self, hparams, *args, hpo_config=None,
                 alg=None, dataset=None, algorithm_generator=None, alg_args=None,
                 alg_kwargs=None, dataset_args=None, dataset_kwargs=None, post_train_hook=None, **kwargs):

        if hpo_config is None:
            hpo_config = HPOConfig(**kwargs)

        super().__init__(*args, hparams=hpo_config)
        logger.info(f"Creating new study (Beam version: {__version__})")

        self.experiment_hparams = hparams

        self.experiment_hparams.set('reload', False)
        self.experiment_hparams.set('override', False)
        self.experiment_hparams.set('print_results', self.hparams.get('print_results'))
        self.experiment_hparams.set('visualize_weights', False)
        self.experiment_hparams.set('enable_tqdm', self.hparams.get('enable_tqdm', False))
        self.experiment_hparams.set('n_gpus', 0)

        exptime = time.strftime("%Y%m%d_%H%M%S", time.localtime())
        self.identifier = f'{self.experiment_hparams.identifier}_hp_optimization_{exptime}'
        self.experiment_hparams.set('identifier', self.identifier)

        self.alg = alg
        if algorithm_generator is None:
            self.ag = partial(beam_algorithm_generator, dataset=dataset,
                              alg_args=alg_args, alg_kwargs=alg_kwargs, dataset_args=dataset_args,
                              dataset_kwargs=dataset_kwargs)
        else:
            self.ag = algorithm_generator
        self.device = beam_device(self.experiment_hparams.device)

        if self.hparams.get('print_hyperparameters'):
            print_beam_hyperparameters(self.experiment_hparams, default_params=HPOConfig(return_defaults=True))

        hpo_path = self.hparams.get('hpo_path')
        if hpo_path is not None:

            root_path = beam_path(hpo_path)
            hpo_path = str(root_path.joinpath(self.experiment_hparams.project_name,
                                              self.experiment_hparams.algorithm,
                                              self.experiment_hparams.identifier))

        else:
            logger.warning("No hpo_path specified. HPO results will be saved only to each experiment directory.")
            root_path = beam_path(self.experiment_hparams.get('logs_path'))
            if type(root_path) is BeamPath:
                hpo_path = str(root_path.joinpath('hpo', self.experiment_hparams.project_name,
                                                  self.experiment_hparams.algorithm,
                                                  self.experiment_hparams.identifier))

        self.hpo_path = hpo_path
        self.experiments_tracker = []
        self.suggestions = {}
        self.post_train_hook = post_train_hook

    @staticmethod
    def get_optimization_mode(mode, objective_name):
        return NeuralAlgorithm.get_optimization_mode(mode, objective_name)

    def add_suggestion(self, param, func, *args, **kwargs):
        self.suggestions[param] = {'func': func, 'args': args, 'kwargs': kwargs}

    def linspace(self, param, start, end, n_steps, endpoint=True,  dtype=None):
        param = param.replace('-', '_').strip()
        self.suggestions[param] = partial(self._linspace, param=param, start=start, end=end, n_steps=n_steps,
                       endpoint=endpoint, dtype=dtype)

    def logspace(self, param, start, end, n_steps, base=None, dtype=None):
        param = param.replace('-', '_').strip()
        self.suggestions[param] = partial(self._logspace, param=param, start=start, end=end,
                                          n_steps=n_steps, base=base, dtype=dtype)

    def uniform(self, param, start, end):
        param = param.replace('-', '_').strip()
        self.suggestions[param] = partial(self._uniform, param=param, start=start, end=end)

    def loguniform(self, param, start, end):
        param = param.replace('-', '_').strip()
        self.suggestions[param] = partial(self._loguniform, param=param, start=start, end=end)

    def categorical(self, param, choices):
        param = param.replace('-', '_').strip()
        self.suggestions[param] = partial(self._categorical, param=param, choices=choices)

    def randn(self, param, mu, sigma):
        param = param.replace('-', '_').strip()
        self.suggestions[param] = partial(self._randn, param=param, mu=mu, sigma=sigma)

    def get_suggestions(self, *args, **kwargs):

        config = {}
        for k, v in self.suggestions.items():
            config[k] = v(*args, **kwargs)

        return config

    @staticmethod
    def _linspace(*args, **kwargs):
        raise NotImplementedError

    @staticmethod
    def _logspace(*args, **kwargs):
        raise NotImplementedError

    @staticmethod
    def _uniform(*args, **kwargs):
        raise NotImplementedError

    @staticmethod
    def _loguniform(*args, **kwargs):
        raise NotImplementedError

    @staticmethod
    def _categorical(*args, **kwargs):
        raise NotImplementedError

    @staticmethod
    def _randn(*args, **kwargs):
        raise NotImplementedError

    def choice(self, param, *args, **kwargs):
        raise NotImplementedError

    def quniform(self, param, *args, **kwargs):
        raise NotImplementedError

    def qloguniform(self, param, *args, **kwargs):
        raise NotImplementedError

    def qrandn(self, param, *args, **kwargs):
        raise NotImplementedError

    def randint(self, param, *args, **kwargs):
        raise NotImplementedError

    def qrandint(self, param, *args, **kwargs):
        raise NotImplementedError

    def lograndint(self, param, *args, **kwargs):
        raise NotImplementedError

    def qlograndint(self, param, *args, **kwargs):
        raise NotImplementedError

    def grid_search(self, param, *args, **kwargs):
        raise NotImplementedError

    def sample_from(self, param, *args, **kwargs):
        raise NotImplementedError

    def discrete_uniform(self, param, *args, **kwargs):
        raise NotImplementedError

    def float(self, param, *args, **kwargs):
        raise NotImplementedError

    def int(self, param, *args, **kwargs):
        raise NotImplementedError

    def tracker(self, algorithm=None, results=None, hparams=None, suggestion=None):

        tracker = {}

        if algorithm is not None and self.hparams.get('track_algorithms'):
            tracker['algorithm'] = algorithm

        if results is not None and self.hparams.get('track_results'):
            tracker['results'] = results

        if hparams is not None and self.hparams.get('track_hparams'):
            tracker['hparams'] = hparams

        if suggestion is not None and self.hparams.get('track_suggestion'):
            tracker['suggestion'] = suggestion

        if len(tracker):
            self.experiments_tracker.append(tracker)

        if self.hpo_path is not None:
            path = beam_path(self.hpo_path).joinpath('tracker')
            path.mkdir(parents=True, exist_ok=True)
            path.joinpath('tracker.pkl').write(tracker)

    def generate_hparams(self, config):

        hparams = copy.deepcopy(self.experiment_hparams)
        if self.experiment_hparams.get('print_hyperparameters'):
            logger.info('Next Hyperparameter suggestion:')
            for k, v in config.items():
                logger.info(k + ': ' + str(v))

        hparams.update(config)
        return hparams

    def runner(self, *args, **kwargs):
        raise NotImplementedError

    def run(self, *args, **kwargs):
        raise NotImplementedError



from ..utils import lazy_property, jupyter_like_traceback
from .resource import beam_llm
from ..core import Processor


class LLMTask(Processor):

    def __init__(self, name=None, description=None, system=None, input_kwargs=None, output_kwargs=None,
                 output_format='json', sep='\n', llm=None, *args, **kwargs):

        super().__init__(*args, name=name, **kwargs)
        self.name = name
        self.description = description
        if input_kwargs is None:
            input_kwargs = {}
        if output_kwargs is None:
            output_kwargs = {}
        self.input_kwargs = input_kwargs
        self.output_kwargs = output_kwargs
        self.output_format = output_format
        self.sep = sep
        self.system = system
        self._llm = llm

    @lazy_property
    def llm(self):
        llm = beam_llm(self._llm)
        return llm

    @llm.setter
    def llm(self, value):
        self.clear_cache('llm')
        self._llm = value

    def prompt(self, **kwargs):
        message = (f"System message: {self.system}\n"
                   f"Task: {self.name}\n"
                   f"Input arguments description: {self.input_kwargs}\n"
                   f"Task description: {self.description}\n"
                   f"Input arguments values: {kwargs}\n"
                   f"Output arguments description: {self.output_kwargs}\n"
                   f"Output format: {self.output_format}\n"
                   f"Your answer should start here:\n\n")
        return message

    def __call__(self, llm_kwargs=None, **kwargs):

        if llm_kwargs is None:
            llm_kwargs = {}
        prompt = self.prompt(**kwargs)

        res = self.llm.ask(prompt, **llm_kwargs)

        try:
            result = res.parse(protocol=self.output_format)
            success = True
        except:
            result = jupyter_like_traceback()
            success = False

        res.add_task_result(result, success=success)

        return res

    def parse(self, response):
        raise NotImplementedError


from typing import List, Union, Any

import numpy as np
import torch

from ..data import BeamData
from ..transformer import Transformer
from ..utils import check_type,lazy_property, as_numpy
from .core import Similarities, BeamSimilarity
from collections import Counter
import scipy.sparse as sp


class ChunkTF(Transformer):

    def __init__(self, *args, sparse_framework='torch', device='cpu', preprocessor=None, **kwargs):
        self.preprocessor = preprocessor or TFIDF.default_preprocessor
        self.sparse_framework = sparse_framework
        self._device = device
        super().__init__(*args, **kwargs)

    @lazy_property
    def device(self):
        from ..utils import beam_device
        return beam_device(self._device)

    @staticmethod
    def tf_tfidf_row(counts, idf=None, scheme='term_frequencies', sparse_framework='torch', norm='l2', k=0.5,
                     log_normalization=False):

        log1p = torch.log1p if sparse_framework == 'torch' else np.log1p
        tf = counts
        if scheme == 'counts_squared':
            tf = tf ** 2
        if scheme in ['term_frequencies', 'raw_counts', 'binary']:
            tf = tf / tf.sum()
        elif scheme == 'double_normalization':
            tf = tf + (1 - k) * tf / tf.max()
        if log_normalization:
            tf = log1p(tf)
        if idf is not None:
            tfidf = tf * idf
        else:
            tfidf = tf

        if norm == 'l2' and sparse_framework == 'torch':
            tfidf = tfidf / torch.norm(tfidf, p=2)
        elif norm == 'l2':
            tfidf = tfidf / np.linalg.norm(tfidf, ord=2)
        elif norm == 'l1' and sparse_framework == 'torch':
            tfidf = tfidf / torch.norm(tfidf, p=1)
        elif norm == 'l1':
            tfidf = tfidf / np.linalg.norm(tfidf, ord=1)

        if scheme == 'raw_counts':
            tf = counts
        elif scheme == 'binary':
            tf = torch.ones_like(counts)

        return tf, tfidf

    def transform_callback(self, x, tokens=None, scheme='term_frequencies', max_token=None, k=0.5, idf=None,
                           norm='l2', log_normalization=False, **kwargs):
        """

        Args:
            k:
            x:
            tokens:
            scheme: can be ['binary', 'term_frequencies', 'log_normalization', 'raw_count', 'double_normalization']
            max_token:
            **kwargs:

        Returns:

        """
        if max_token is None:
            max_token = np.max(list(tokens))

        tfidf = []
        tf = []
        ind_ptrs = [0]
        cols = []

        for xi in x:
            xi = self.preprocessor(xi)
            if scheme == 'binary':
                xi = list(set(xi))
            c = Counter(xi)
            c = Counter({k: v for k, v in c.items() if k in tokens})

            ind_ptrs.append(len(c))
            if self.sparse_framework == 'torch':
                ind = torch.tensor(list(c.keys()), dtype=torch.int64, device=self.device)
                counts = torch.tensor(list(c.values()), dtype=torch.float32, device=self.device)
                tfi, tfidfi = self.tf_tfidf_row(counts, idf=idf[ind], scheme=scheme, sparse_framework=self.sparse_framework,
                                            norm=norm, k=k, log_normalization=log_normalization)

                cols.append(ind)
                tf.append(tfi)
                tfidf.append(tfidfi)

            else:
                ind = np.array(list(c.keys()))
                counts = np.array(list(c.values()))
                tfi, tfidfi = self.tf_tfidf_row(counts, idf=idf[ind], scheme=scheme, sparse_framework=self.sparse_framework,
                               norm=norm, k=k, log_normalization=log_normalization)

                cols.append(ind)
                tf.append(tfi)
                tfidf.append(tfidfi)

        if self.sparse_framework == 'torch':
            ind_ptrs = torch.cumsum(torch.tensor(ind_ptrs, dtype=torch.int64, device=self.device), dim=0)
            tf = torch.sparse_csr_tensor(ind_ptrs, torch.cat(cols), torch.cat(tf), device=self.device,
                                        size=(len(x), max_token + 1))
            tfidf = torch.sparse_csr_tensor(ind_ptrs, torch.cat(cols), torch.cat(tfidf), device=self.device,
                                         size=(len(x), max_token + 1))
        else:
            ind_ptrs = np.cumsum(ind_ptrs)
            tf = sp.csr_matrix((np.concatenate(tf), np.concatenate(cols), ind_ptrs),
                              shape=(len(x), max_token + 1))
            tfidf = sp.csr_matrix((np.concatenate(tfidf), np.concatenate(cols), ind_ptrs),
                               shape=(len(x), max_token + 1))

        return tf, tfidf


class ChunkDF(Transformer):

        def __init__(self, *args, preprocessor=None, **kwargs):
            self.preprocessor = preprocessor or TFIDF.default_preprocessor
            super().__init__(*args, **kwargs)

        def transform_callback(self, x, key=None, is_chunk=False, fit=False, path=None, **kwargs):
            y = Counter()
            y_sum = Counter()
            for xi in x:
                xi = self.preprocessor(xi)
                y.update(set(xi))
                y_sum.update(xi)
            return y, y_sum


class TFIDF(BeamSimilarity):

    def __init__(self, *args, preprocessor=None, min_df=None, max_df=None, max_features=None, use_idf=True,
                 smooth_idf=True, sublinear_tf=False, n_workers=0, mp_method='joblib', chunksize=None,
                 n_chunks=None, sparse_framework='torch', device='cpu', norm='l2',
                 metric='bm25', bm25_k1=1.5, bm25_b=0.75, bm25_epsilon=0.25, **kwargs):

        super().__init__(*args, min_df=min_df, max_df=max_df, max_features=max_features, use_idf=use_idf,
                         sparse_framework=sparse_framework, smooth_idf=smooth_idf, sublinear_tf=sublinear_tf,
                         chunksize=chunksize, n_chunks=n_chunks, device=device, metric=metric,
                         bm25_k1=bm25_k1, bm25_b=bm25_b, bm25_epsilon=bm25_epsilon, norm=norm,
                         n_workers=n_workers, mp_method=mp_method,
                         **kwargs)
        self.df = None
        self.cf = None  # corpus frequency
        self.n_docs = None
        self.tf = None
        self.index = None
        self._is_trained = None
        self.reset()

        self.preprocessor = preprocessor or TFIDF.default_preprocessor

        self.min_df = self.get_hparam('min_df', min_df)
        self.max_df = self.get_hparam('max_df', max_df)
        self.max_features = self.get_hparam('max_features', max_features)
        self.use_idf = self.get_hparam('use_idf', use_idf)
        self.smooth_idf = self.get_hparam('smooth_idf', smooth_idf)
        self.sublinear_tf = self.get_hparam('sublinear_tf', sublinear_tf)
        self.sparse_framework = self.get_hparam('sparse_framework', sparse_framework)
        self.norm = self.get_hparam('norm', norm)
        self.bm25_k1 = self.get_hparam('bm25_k1', bm25_k1)
        self.bm25_b = self.get_hparam('bm25_b', bm25_b)
        self.bm25_epsilon = self.get_hparam('bm25_epsilon', bm25_epsilon)

        # we choose the csr layout for the sparse matrix
        # according to chatgpt it has some advantages over coo:
        # see https://chat.openai.com/share/9028c9f3-9695-4914-a15c-89902efa8837

        self.device = self.get_hparam('device', device)

        self.n_workers = self.get_hparam('n_workers', n_workers)
        n_chunks = self.get_hparam('n_chunks', self.n_workers)
        chunksize = self.get_hparam('chunksize', None)
        mp_method = self.get_hparam('mp_method', mp_method)

        self.chunk_tf = ChunkTF(n_workers=self.n_workers, n_chunks=n_chunks, chunksize=chunksize, mp_method=mp_method,
                                squeeze=False, reduce=False, sparse_framework=self.sparse_framework, device=self.device,
                                preprocessor=self.preprocessor)
        self.chunk_df = ChunkDF(n_workers=self.n_workers, n_chunks=n_chunks, chunksize=chunksize, mp_method=mp_method,
                                squeeze=False, reduce=False, preprocessor=self.preprocessor)

    @staticmethod
    def default_preprocessor(x):

        x_type = check_type(x)

        if x_type.minor == 'torch':
            x = as_numpy(x).tolist()
        elif x_type.minor == 'numpy':
            x = x.tolist()
        else:
            x = list(x)

        return x

    def vstack_csr_tensors(self, x):
        crow_indices = []
        col_indices = []
        values = []
        n = 0
        for xi in x:
            indptr = xi.crow_indices()
            if len(crow_indices) > 0:
                crow_indices.append(indptr[1:] + crow_indices[-1][-1])
            else:
                crow_indices.append(indptr)
            col_indices.append(xi.col_indices())
            values.append(xi.values())
            n += len(xi)
            
        return torch.sparse_csr_tensor(torch.cat(crow_indices), torch.cat(col_indices), torch.cat(values),
                                       size=(n, xi.shape[-1]), device=self.device)

    def tf_and_tfidf(self, x, tokens=None, scheme=None, idf=None, norm=None, k=.5, log_normalization=None, **kwargs):

        idf = self.idf if idf is None else idf
        norm = self.norm if norm is None else norm
        log_normalization = self.sublinear_tf if log_normalization is None else log_normalization

        if tokens is None:
            tokens = self.tokens
            max_token = self.max_token
        else:
            max_token = max(tokens)

        chunks = self.chunk_tf.transform(x, tokens=tokens, max_token=max_token, scheme=scheme, idf=idf, k=k,
                                         norm=norm, log_normalization=log_normalization, **kwargs)

        if self.sparse_framework == 'torch':
            tf = self.vstack_csr_tensors([c[0] for c in chunks])
            tfidf = self.vstack_csr_tensors([c[1] for c in chunks])
        else:
            tf = sp.vstack([c[0] for c in chunks])
            tfidf = sp.vstack([c[1] for c in chunks])

        return tf, tfidf

    def bm25(self, q, k1=1.5, b=0.75, epsilon=.25):

        idf = self.idf_bm25(epsilon=epsilon)
        _, q_tfidf = self.tf_and_tfidf(q, scheme='counts_times_idf', idf=idf, norm='none', log_normalization=False)

        if self.sparse_framework == 'torch':
            len_norm_values = (1 - b) + (b / self.avg_doc_len) * self.doc_len_sparse.values()
            bm25_tf_values = self.tf.values() * (k1 + 1) / (self.tf.values() + k1 * len_norm_values)
            bm25_tf = torch.sparse_csr_tensor(self.tf.crow_indices(), self.tf.col_indices(), bm25_tf_values,
                                              size=self.tf.shape, device=self.device)
        else:
            len_norm_values = (1 - b) + (b / self.avg_doc_len) * self.doc_len_sparse.data
            bm25_tf_values = self.tf.data * (k1 + 1) / (self.tf.data + k1 * len_norm_values)
            bm25_tf = sp.csr_matrix((bm25_tf_values, self.tf.indices, self.tf.indptr), shape=self.tf.shape)

        if self.sparse_framework == 'torch':
            scores = torch.matmul(bm25_tf, q_tfidf.to_dense().T).T
        else:
            q_idf = q_tfidf.multiply(idf)
            scores = q_idf @ bm25_tf.T

        return scores

    def transform(self, x: Union[List, List[List], BeamData], index: Union[None, Any] = None,
                  add_to_index: bool = False):

        x, index = self.extract_data_and_index(x, index)
        if add_to_index or self.index is None:
            self.add_index(x, index)

        x_type = check_type(x)
        if not x_type.major == 'container':
            x = [x]

        tf, tfidf = self.tf_and_tfidf(x, scheme='raw_counts')
        if self.tf is None:
            self.tf = tf
        else:
            if self.sparse_framework == 'torch':
                self.tf = self.vstack_csr_tensors([self.tf, tf])
            else:
                self.tf = sp.vstack([self.tf, tf])

        return tfidf

    def reset(self):
        self.df = Counter()
        self.cf = Counter()
        self.n_docs = 0
        self.tf = None
        self.index = None
        self._is_trained = False
        self.clear_cache('idf', 'tokens', 'n_tokens', 'avg_doc_len', 'idf_bm25', 'doc_len', 'doc_len_sparse',
                         'max_token')

    @lazy_property
    def tokens(self):
        """Build a mapping from tokens to indices based on filtered tokens."""
        return set(self.df.keys())

    @lazy_property
    def avg_doc_len(self):
        return sum(self.cf.values()) / self.n_docs

    @lazy_property
    def n_tokens(self):
        return max(list(self.tokens))

    @lazy_property
    def idf(self):

        if self.use_idf:
            if self.smooth_idf:
                idf_version = 'smooth'
            else:
                idf_version = 'standard'
        else:
            idf_version = 'unary'

        return self.calculate_idf(scheme=idf_version)

    def idf_bm25(self, epsilon=.25):
        return self.calculate_idf(scheme='bm25', epsilon=epsilon)

    @lazy_property
    def max_token(self):
        return max(list(self.tokens))

    @lazy_property
    def doc_len(self):
        if self.sparse_framework == 'torch':
            doc_lengths = self.tf.sum(dim=1, keepdim=True).to_dense().squeeze(-1)
        else:
            doc_lengths = np.array(self.tf.sum(axis=1))

        return doc_lengths

    @lazy_property
    def doc_len_sparse(self):
        if self.sparse_framework == 'torch':
            repeats = self.tf.crow_indices().diff()
            values = torch.repeat_interleave(self.doc_len, repeats, dim=0)
            return torch.sparse_csr_tensor(self.tf.crow_indices(), self.tf.col_indices(), values,
                                           size=self.tf.shape, device=self.device)
        else:
            repeats = np.diff(self.tf.indptr)
            values = np.repeat(self.doc_len, repeats)
            return sp.csr_matrix((values, self.tf.indices, self.tf.indptr), shape=self.tf.shape)

    def calculate_idf(self, scheme='standard', epsilon=.25, okapi=True):
        """Calculate the inverse document frequency (IDF) vector.
        version: str, default='standard' [standard, smooth, unary, bm25]
        """
        n_docs = self.n_docs
        keys = list(self.df.keys())

        if self.sparse_framework == 'torch':
            col_indices = torch.tensor(keys, dtype=torch.int64, device=self.device)
            log = torch.log
            zeros = torch.zeros
            framework_kwargs = {'device': self.device, 'dtype': torch.float32}
            array = torch.tensor
        else:
            col_indices = np.array(keys)
            log = np.log
            zeros = np.zeros
            framework_kwargs = {}
            array = np.array

        vals = array(list(self.df.values()), **framework_kwargs)
        nq = zeros(col_indices.max()+1, **framework_kwargs)

        if scheme == 'unary':
            nq[col_indices] = 1
        else:
            nq[col_indices] = vals

        if scheme == 'standard':
            idf = log(n_docs / nq)
        elif scheme == 'smooth':
            idf = log(n_docs / (nq + 1)) + 1
        elif scheme == 'bm25':
            idf = log((n_docs - nq + .5) / (nq + .5) + (1 - int(okapi)))
            idf[idf < 0] = epsilon * idf.mean()
        else:
            raise ValueError(f"Unknown version: {scheme}")

        return idf

    def fit(self, x=None, **kwargs):
        if x is not None:
            self.reset()
            self.add(x, **kwargs)
        self.filter_tokens()
        self._is_trained = True

    def fit_transform(self, x, index=None, **kwargs):

        self.fit(x, **kwargs)
        return self.transform(x, index=index, add_to_index=True)

    def add(self, x, **kwargs):
        self.n_docs += len(x)
        chunks = self.chunk_df.transform(x, **kwargs)
        for df, cf in chunks:
            self.df.update(df)
            self.cf.update(cf)

    def filter_tokens(self):

        n = self.n_docs
        if self.min_df is not None:
            min_df = self.min_df
            if self.min_df < 1:
                min_df = int(self.min_df * n)
            self.df = {k: v for k, v in self.df.items() if v >= min_df}

        if self.max_df is not None:
            max_df = self.max_df
            if self.max_df < 1:
                max_df = int(self.max_df * n)
            self.df = {k: v for k, v in self.df.items() if v <= max_df}

        if self.max_features is not None:
            self.df = Counter(dict(sorted(self.df.items(), key=lambda x: x[1], reverse=True)[:self.max_features]))

        self.cf = {k: v for k, v in self.cf.items() if k in self.df}

    def train(self, x):
        self.fit(x)

    def search(self, q, k=1, **kwargs):

        if self.metric == 'bm25':
            scores = self.bm25(q, **kwargs)
            if self.sparse_framework == 'torch':
                topk = torch.topk(scores, k=k, dim=1)
                I = topk.indices
                D = topk.values
            else:
                I = scores.argsort(axis=1)[:, -k:]
                D = scores[np.arange(len(scores)), I]
        else:
            raise ValueError(f"Unknown metric: {self.metric}")

        return Similarities(index=self.get_index(I), distance=D, metric=self.metric, model='tfidf')

    @property
    def exclude_pickle_attributes(self):
        return ['df', 'cf', 'n_docs', 'tf', 'index']




import sys
import loguru
from ..utils import running_platform
from ..path import beam_path
import atexit


class BeamLogger:

    def __init__(self, path=None, print=True):
        self.logger = loguru.logger
        self.logger.remove()
        self.handlers_queue = []
        self.running_platform = running_platform()

        self.handlers = {}
        if print:
            self.print()

        self.file_objects = {}
        self.path = None
        if path is not None:
            self.add_file_handlers(path)

        atexit.register(self.cleanup)

    def dont_print(self):
        self.logger.remove(self.handlers['stdout'])

    def print(self):
        self.handlers['stdout'] = self.stdout_handler()

    def cleanup(self, print=True):
        for k, handler in self.handlers.items():
            if k == 'stdout' and print:
                continue
            try:
                self.logger.remove(handler)
            except ValueError:
                pass

        if print:
            self.handlers = {k: v for k, v in self.handlers.items() if k == 'stdout'}
        else:
            self.handlers = {}

        for k, file_object in self.file_objects.items():
            file_object.close()
        self.file_objects = {}

    def add_file_handlers(self, path):

        path = beam_path(path)

        debug_path = path.joinpath('debug.log')
        file_object = debug_path.open('w')
        self.file_objects[str(debug_path)] = file_object

        if self.running_platform == 'script':
            format = '{time:YYYY-MM-DD HH:mm:ss} ({elapsed}) | BeamLog | {level} | {file} | {function} | {line} | {message}'
        else:
            format = '{time:YYYY-MM-DD HH:mm:ss} ({elapsed}) | BeamLog | {level} | %s | {function} | {line} | {message}' \
                     % self.running_platform

        handler = self.logger.add(file_object, level='DEBUG', format=format)

        self.handlers[str(debug_path)] = handler

        json_path = path.joinpath('json.log')
        file_object = json_path.open('w')
        self.file_objects[str(json_path)] = file_object

        format = 'JSON LOGGER'
        handler = self.logger.add(file_object, level='DEBUG', format=format, serialize=True)

        self.handlers[str(json_path)] = handler

    def open(self, path):
        path = beam_path(path)
        self.handlers_queue.append(path)
        return self

    def __enter__(self):
        self.add_file_handlers(self.handlers_queue[-1])
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        path = self.handlers_queue.pop()
        self.remove_file_handler(path)

    def remove_file_handler(self, name):
        for suffix in ['debug.log', 'json.log']:
            fullname = str(name.joinpath(suffix))
            self.logger.remove(self.handlers[fullname])
            self.handlers.pop(fullname)

    def debug(self, message, **extra):
        self.logger.debug(message, **extra)

    def info(self, message, **extra):
        self.logger.info(message, **extra)

    def warning(self, message, **extra):
        self.logger.warning(message, **extra)

    def error(self, message, **extra):
        self.logger.error(message, **extra)

    def critical(self, message, **extra):
        self.logger.critical(message, **extra)

    def exception(self, message, **extra):
        self.logger.exception(message, **extra)

    def __getstate__(self):

        if self.path is None:
            path = None
        elif isinstance(self.path, str):
            path = self.path
        else:
            path = self.path.as_uri()
        state = {'path': path}
        return state

    def __setstate__(self, state):
        self.__init__(state['path'])

    def stdout_handler(self, level='INFO'):
        return self.logger.add(sys.stdout, level=level, colorize=True, format=
        '<green>{time:YYYY-MM-DD HH:mm:ss}</green> | BeamLog | <level>{level}</level> | <level>{message}</level>')

    def set_verbosity(self, level):
        """
        Sets the log level for all handlers to the specified level.
        """
        # Convert the level string to uppercase to match Loguru's expected levels
        level = level.upper()

        if 'stdout' in self.handlers:
            self.logger.remove(self.handlers['stdout'])

        self.handlers['stdout'] = self.stdout_handler(level=level)


beam_logger = BeamLogger()



import numpy as np
import torch
import sys
import itertools
import pandas as pd
import math
import hashlib

from ..logger import beam_logger as logger
from ..utils import as_tensor, check_type, as_numpy, beam_device


class UniversalBatchSampler(object):
    """
         A class used to generate batches of indices, to be used in drawing samples from a dataset
         ...
         Attributes
         ----------
         indices : tensor
             The array of indices that can be sampled.
         length : int
               Maximum number of batches that can be returned by the sampler
         size : int
               The length of indices
         batch_size: int
               size of batch
         minibatches : int
             number of batches in one iteration over the array of indices
         once : bool
             If true, perform only 1 iteration over the indices array.
         tail : bool
             If true, run over the tail elements of indices array (the remainder left
             when dividing len(indices) by batch size). If once, return a minibatch. Else
             sample elements from the rest of the array to supplement the tail elements.
          shuffle : bool
             If true, shuffle the indices after each epoch
         """

    def __init__(self, indices, batch_size, probs=None, length=None, shuffle=True, tail=True,
                 once=False, expansion_size=int(1e7), dynamic=False, buffer_size=None,
                 probs_normalization='sum', sample_size=100000, device=None):

        """
               Parameters
               ----------
               indices : array/tensor/int
                   If array or tensor, represents the indices of the examples contained in a subset of the whole data
                   (train/validation/test). If int, generates an array of indices [0, ..., dataset_size].
               batch_size : int
                   number of elements in a batch
               probs : array, optional
                   An array the length of indices, with probability/"importance" values to determine
                   how to perform oversampling (duplication of indices to change data distribution).
               length : int, optional
                  see descrtiption in class docstring
               shuffle : bool, optional
                  see description in class docstring
               tail : bool, optional
                  see description in class docstring
               once: bool, optional
                  see description in class docstring
               expansion_size : int
                    Limit on the length of indices (when oversampling, the final result can't be longer than
                    expansion_size).``
         """

        self.length = sys.maxsize if length is None else int(length)
        self.once = once
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.tail = tail
        self.probs_normalization = probs_normalization
        self.buffer_size = buffer_size
        self.refreshed = False
        self.size = None
        self.minibatches = None
        self.sample_size = sample_size

        indices_type = check_type(indices)
        if indices_type.minor == 'torch':
            device = device if device is not None else indices.device
        else:
            device = device if device is not None else 'cpu'\

        self.device = beam_device(device)

        if indices_type.major == 'array':
            self.indices = as_tensor(indices, device=device, dtype=torch.int64)
        else:
            self.indices = torch.arange(indices, device=device)
        self.probs = as_numpy(probs) if probs is not None else None

        if dynamic:
            self.samples_iterator = self.dynamic_samples_iterator
            self.indices = as_numpy(self.indices)

        else:
            self.samples_iterator = self.static_samples_iterator
            if probs is not None:

                logger.info("UniversalBatchSampler: Building expanded indices array based on given probabilities")
                probs = as_numpy(self.normalize_probabilities(probs))
                grow_factor = max(expansion_size, len(probs)) / len(probs)

                probs = (probs * len(probs) * grow_factor).round().astype(np.int)
                m = np.gcd.reduce(probs)
                reps = np.clip(np.round(probs / m).astype(np.int), 1, None)

                logger.info(f"Expansion size: {expansion_size}, before expansion: {len(probs)}, "
                            f"after expansion: {np.sum(reps)}")
                indices = pd.DataFrame({'index': as_numpy(self.indices), 'times': reps})
                self.indices = as_tensor(indices.loc[indices.index.repeat(indices['times'])]['index'].values,
                                         device=self.device, dtype=torch.int64)

        self.size = len(self.indices)
        self.minibatches = int(self.size / self.batch_size)
        if once:
            self.length = math.ceil(self.size / batch_size) if tail else self.size // batch_size

    def normalize_probabilities(self, p):

        if p is None:
            return None

        if self.probs_normalization == 'softmax':
            return torch.softmax(as_tensor(p, device='cpu'), dim=0)

        return p / p.sum()

    def update_fifo(self):
        if self.buffer_size is not None:
            self.indices = self.indices[-self.buffer_size:]
            self.probs = self.probs[-self.buffer_size:]
            self.unnormalized_probs = self.unnormalized_probs[-self.buffer_size:]

    def dynamic_samples_iterator(self):

        self.n = 0
        for _ in itertools.count():

            self.update_fifo()
            probs = as_numpy(self.normalize_probabilities(self.probs))
            size = min(self.size, self.sample_size) if self.sample_size is not None else self.size
            minibatches = math.ceil(size / self.batch_size)
            indices_batched = torch.LongTensor(np.random.choice(self.indices, size=(minibatches, self.batch_size),
                                                        replace=True, p=probs))

            for samples in indices_batched:
                self.n += 1
                yield samples
                if self.n >= self.length:
                    return
                if self.refreshed:
                    self.refreshed = False
                    continue

    def replace_indices(self, indices, probs=None):
        if check_type(indices).major == 'array':
            self.indices = as_numpy(indices)
        else:
            self.indices = np.arange(indices)
        self.probs = as_numpy(probs) if probs is not None else None
        self.refreshed = True

    def append_indices(self, indices, probs=None):
        self.indices = np.concatenate([self.indices, as_numpy(indices)])
        if probs is not None:
            self.probs = torch.cat([self.probs, as_tensor(probs, device='cpu')])

    def append_index(self, index, prob=None):
        self.indices = np.concatenate([self.indices, as_numpy([index])])
        if prob is not None:
            self.probs = torch.cat([self.probs, as_tensor([prob], device='cpu')])

    def pop_index(self, index):
        v = self.indices != index
        self.indices = self.indices[v]
        if self.probs is not None:
            self.probs = self.probs[torch.BoolTensor(v)]

    def pop_indices(self, indices):
        v = ~np.isin(self.indices, as_numpy(indices))
        self.indices = self.indices[v]
        if self.probs is not None:
            self.probs = self.probs[v]

    def static_samples_iterator(self):

        self.n = 0
        indices = self.indices.clone()

        for _ in itertools.count():

            if self.shuffle:
                indices = indices[torch.randperm(len(indices), device=self.device)]

            indices_batched = indices[:self.minibatches * self.batch_size]
            indices_tail = indices[self.minibatches * self.batch_size:]

            if self.tail and not self.once:

                to_sample = max(0, self.batch_size - (self.size - self.minibatches * self.batch_size))

                try:
                    fill_batch = np.random.choice(len(indices_batched), to_sample, replace=(to_sample > self.size))
                except ValueError:
                    raise ValueError("Looks like your dataset is smaller than a single batch. Try to make it larger.")

                fill_batch = indices_batched[as_tensor(fill_batch, device=self.device)]
                indices_tail = torch.cat([indices_tail, fill_batch])

                indices_batched = torch.cat([indices_batched, indices_tail])

            indices_batched = indices_batched.reshape((-1, self.batch_size))

            for samples in indices_batched:
                self.n += 1
                yield samples
                if self.n >= self.length:
                    return

            if self.once:
                if self.tail:
                    yield indices_tail
                return

    def __iter__(self):
        return self.samples_iterator()

    def __len__(self):
        return self.length


class HashSplit(object):

    def __init__(self, seed=None, granularity=.001, **argv):

        s = pd.Series(index=list(argv.keys()), data=list(argv.values()))
        s = s / s.sum() / granularity
        self.subsets = s.cumsum()
        self.n = int(1 / granularity)
        self.seed = seed

    def __call__(self, x):

        if type(x) is pd.Series:
            return x.apply(self._call)
        elif type(x) is list:
            return [self._call(xi) for xi in x]
        else:
            return self._call(x)

    def _call(self, x):

        x = f'{x}/{self.seed}'
        x = int(hashlib.sha1(x.encode('utf-8')).hexdigest(), 16) % self.n
        subset = self.subsets.index[x < self.subsets][0]

        return subset



import inspect
import re
import time
import uuid

try:
    from collections.abc import Iterator
except ImportError:
    # assume python < 3.10
    from collections import Iterator

from ..logger import beam_logger as logger
from ..utils import parse_text_to_protocol, retry


class LLMResponse:
    def __init__(self, response, llm, prompt=None, prompt_kwargs=None, chat=False, stream=False, parse_retrials=3,
                 sleep=1, prompt_type='completion', **kwargs):
        self.response = response
        self._prompt = prompt
        self._prompt_kwargs = prompt_kwargs
        self.parse_retrials = parse_retrials
        self.sleep = sleep
        self.llm = llm
        self.id = f'beamllm-{uuid.uuid4()}'
        self.model = llm.model
        self.created = int(time.time())
        self.chat = chat

        if stream:
            self.object = "chat.completion.chunk"
        elif chat:
            self.object = "chat.completion"
        else:
            self.object = "text.completion"

        self.stream = stream
        self.prompt_type = prompt_type
        self._task_result = None
        self._task_success = None

        if not inspect.isgenerator(self.response) and not isinstance(self.response, Iterator):
            assert self.verify(), "Response is not valid"

    def __str__(self):
        return self.text

    def __bool__(self):
        return self.bool

    def __int__(self):
        return self.int

    def __float__(self):
        return self.float

    def __iter__(self):
        if not self.stream:
            yield self
        else:
            for r in self.response:
                yield LLMResponse(r, self.llm, prompt=self.prompt, chat=self.chat, stream=self.stream,
                                  prompt_kwargs=self._prompt_kwargs, parse_retrials=self.parse_retrials,
                                  sleep=self.sleep, prompt_type=self.prompt_type)

    def add_task_result(self, task_result, success=True):
        self._task_result = task_result
        self._task_success = success

    @property
    def int(self):
        try:
            return int(self.text)
        except:
            return None

    @property
    def float(self):
        try:
            return float(self.text)
        except:
            return None

    @property
    def task_result(self):
        return self._task_result

    @property
    def task_success(self):
        return self._task_success

    @property
    def prompt(self):
        return self._prompt

    @property
    def prompt_kwargs(self):
        return self._prompt_kwargs

    def verify(self):
        return self.llm.verify_response(self)

    @property
    def text(self):
        return self.llm.extract_text(self)

    @property
    def openai_format(self):
        return self.llm.openai_format(self)

    def parse_text(self, text, protocol='json'):
        return self._protocol(text, protocol=protocol)

    def parse(self, protocol='json'):
        return self._protocol(self.text, protocol=protocol)

    def _protocol(self, text, protocol='json'):

        if self.parse_retrials == 0:
            return parse_text_to_protocol(text, protocol=protocol)
        try:
            return parse_text_to_protocol(text, protocol=protocol)
        except:
            retry_protocol = (retry(retrials=self.parse_retrials, sleep=self.sleep,  logger=logger, name=f"fix-{protocol} with {self.model}")
                              (self.llm.fix_protocol))
            return retry_protocol(text, protocol=protocol)

    @property
    def judge(self):
        # boolean judgement of response by LLM for the prompt
        # return self.llm.judge(self.prompt, self.text, **self.prompt_kwargs)
        return self.llm.judge(self.prompt, self.text)

    @property
    def bool(self):

        text = re.findall(r'\w+', self.text.lower())
        n_words = len(text)

        if 'true' in text:
            return True
        elif 'false' in text:
            return False
        elif 'yes' in text:
            return True
        elif 'no' in text and n_words == 1:
            return False
        return None

    @property
    def json(self):
        json_text = self.llm.extract_text(self)
        json_text = json_text.replace(r'/_', '_')
        json_text = json_text.replace('False', 'false')
        json_text = json_text.replace('True', 'true')
        return self._protocol(json_text, protocol='json')

    @property
    def html(self):
        text = self.llm.extract_text(self)
        return self._protocol(text, protocol='html')

    @property
    def xml(self):
        text = self.llm.extract_text(self)
        return self._protocol(text, protocol='xml')

    @property
    def csv(self):
        text = self.llm.extract_text(self)
        return self._protocol(text, protocol='csv')

    @property
    def yaml(self):
        text = self.llm.extract_text(self)
        # text = re.search('yaml\n([\s\S]*?)\n', text).group(1)
        return self._protocol(text, protocol='yaml')

    @property
    def toml(self):
        text = self.llm.extract_text(self)
        return self._protocol(text, protocol='toml')

    @property
    def choices(self):
        return self.llm.extract_choices(self.response)



import json
from collections import OrderedDict
import inspect
import pickle
from argparse import Namespace
import io
from ..path import beam_path, normalize_host
from ..utils import retrieve_name, lazy_property, check_type
from ..config import BeamConfig


class MetaBeamInit(type):
    def __call__(cls, *args, _store_init_path=None, **kwargs):
        init_args = {'args': args, 'kwargs': kwargs}
        if _store_init_path:
            cls._pre_init(_store_init_path, init_args)
        instance = super().__call__(*args, **kwargs)
        instance._init_args = init_args
        instance._init_is_done = True
        return instance

    def _pre_init(cls, store_init_path, init_args):
        # Process or store arguments
        store_init_path = beam_path(store_init_path)
        store_init_path.write(init_args, ext='.pkl')


class BeamBase(metaclass=MetaBeamInit):

    def __init__(self, *args, name=None, **kwargs):

        self._init_is_done = False
        self._name = name
        self._lazy_cache = {}

    def getattr(self, attr):
        raise AttributeError(f"Attribute {attr} not found")

    def __getattr__(self, item):
        if (item.startswith('_') or item == '_init_is_done' or not hasattr(self, '_init_is_done')
                or not self._init_is_done):
            return object.__getattribute__(self, item)
        return self.getattr(item)

    def clear_cache(self, *args):
        if len(args) == 0:
            self._lazy_cache = {}
        else:
            for k in args:
                if k in self._lazy_cache:
                    del self._lazy_cache[k]

    def in_cache(self, attr):
        return attr in self._lazy_cache

    @property
    def name(self):
        if self._name is None and hasattr(self, '_init_is_done') and self._init_is_done:
            self._name = retrieve_name(self)
        return self._name

    def beam_class(self):
        return self.__class__.__name__


class Processor(BeamBase):

    skeleton_file = 'skeleton'
    state_file = 'state'

    def __init__(self, *args, name=None, hparams=None, override=True, remote=None, llm=None, **kwargs):

        super().__init__(name=name)
        self.remote = remote

        if len(args) > 0:
            self.hparams = args[0]
        elif hparams is not None:
            self.hparams = hparams
        else:
            if not hasattr(self, 'hparams'):
                self.hparams = BeamConfig(config=Namespace())

        for k, v in kwargs.items():
            v_type = check_type(v)
            if v_type.major in ['scalar', 'none']:
                if k not in self.hparams or override:
                    self.hparams[k] = v

        self._llm = self.get_hparam('llm', llm)

    @lazy_property
    def llm(self):
        if type(self._llm) is str:
            from ..resource import resource
            self._llm = resource(self._llm)
        return self._llm

    @property
    def exclude_pickle_attributes(self):
        '''
        return of list of class attributes that are used to save the state and are not part of the
        skeleton of the instance. override this function to add more attributes to the state and avoid pickling a large
        skeleton.
        @return:
        '''
        keys = self.state_dict().keys()
        keys = [k for k in keys if k not in ['hparams']]

        return keys

    def __getstate__(self):
        # Create a new state dictionary with only the skeleton attributes without the state attributes
        state = {k: v for k, v in self.__dict__.items() if k not in self.exclude_pickle_attributes}
        return state

    def __setstate__(self, state):
        # Restore the skeleton attributes
        self.__dict__.update(state)

    @classmethod
    def from_remote(cls, hostname, *args, port=None,  **kwargs):

        hostname = normalize_host(hostname, port=port)
        from ..serve.client import BeamClient
        remote = BeamClient(hostname)
        self = cls(*args, remote=remote, **kwargs)

        def detour(self, attr):
            return getattr(self.remote, attr)

        setattr(self, '__getattribute__', detour)

        return self

    @classmethod
    def from_arguments(cls, *args, **kwargs):
        return cls(*args, **kwargs)

    @classmethod
    def from_path(cls, path):
        path = beam_path(path)

        skeleton_file = list(path.glob(f"{Processor.skeleton_file}*"))[0]
        state_file = list(path.glob(f"{Processor.state_file}*"))[0]

        obj = skeleton_file.read()
        obj.load_state(state_file)

        return obj

    @classmethod
    def from_state_path(cls, path):
        path = beam_path(path)
        state = path.read()
        kwargs = dict()
        args = tuple()
        if 'aux' in state:
            if 'kwargs' in state['aux']:
                kwargs = state['aux']['kwargs']
            if 'args' in state['aux']:
                args = state['aux']['args']
        hparams = BeamConfig(config=state['hparams'])
        alg = cls(hparams, *args, **kwargs)
        alg.load_state(state)
        return alg

    @classmethod
    def from_nlp(cls, query, llm=None, ask_kwargs=None, **kwargs):
        from ..resource import resource
        from ..logger import beam_logger as logger

        llm = resource(llm)

        def is_class_method(member):
            # First, ensure that member is a method bound to a class
            if inspect.ismethod(member) and inspect.isclass(member.__self__):
                # Now that we've confirmed member is a method, check the name conditions
                if not member.__name__.startswith('_') and member.__name__ != 'from_nlp':
                    return True
            return False

        classmethods = [name for name, member in inspect.getmembers(cls, predicate=is_class_method)]

        example_output = {'method': 'method_name'}
        prompt = (f"Choose the suitable classmethod that should be used to build a class instance according to the "
                  f"following query:\n"
                  f"Query: {query}\n"
                  f"Class: {cls.__name__}\n"
                  f"Methods: {classmethods}\n"
                  f"Return your answer as a JSON object of the following form:\n"
                  f"{json.dumps(example_output)}\n"
                  f"Your answer:\n\n")

        ask_kwargs = ask_kwargs or {}
        response = llm.ask(prompt, **ask_kwargs).json

        constructor_name = response['method']

        if constructor_name not in classmethods:
            raise ValueError(f"Constructor {constructor_name} not found in the list of class constructors")

        constructor = getattr(cls, constructor_name)
        logger.info(f"Using classmethod {constructor_name} to build the class instance")

        constructor_sourcecode = inspect.getsource(constructor)
        init_sourcecode = inspect.getsource(cls.__init__)

        json_output_example = {"args": ['arg1', 'arg2'], "kwargs": {'kwarg1': 'value1', 'kwarg2': 'value2'}}
        prompt = (f"Build a suitable dictionary of arguments and keyword arguments to build a class instance according "
                  f"to the following query:\n"
                  f"Query: {query}\n"
                  f"with the classmethod: {constructor_name} (of class {cls.__name__}) with source-code:\n"
                  f"{constructor_sourcecode}\n"
                  f"and the class __init__ method source-code:\n"
                  f"{init_sourcecode}\n"
                  f"Return your answer as a JSON object of the following form:\n"
                  f"{json_output_example}\n"
                  f"Your answer:\n\n")

        d = llm.ask(prompt, **ask_kwargs).json
        args = d.get('args', [])
        kwargs = d.get('kwargs', {})

        logger.info(f"Using args: {args} and kwargs: {kwargs} to build the class instance")

        return constructor(*args, **kwargs)

    def get_hparam(self, hparam, default=None, preferred=None, specific=None):
        return self.hparams.get(hparam, default=default, preferred=preferred, specific=specific)

    def set_hparam(self, hparam, value, tags=None):
        self.hparams.set(hparam, value, tags=tags)

    def update_hparams(self, hparams, tags=None):
        self.hparams.update(hparams, tags=tags)

    def to_bundle(self, path):
        from ..auto import AutoBeam
        AutoBeam.to_bundle(self, path)

    def state_dict(self):
        # The state must contain a key 'hparams' with the hparams of the instance
        return {'hparams': self.hparams}

    def load_state_dict(self, state_dict):
        for k, v in state_dict.items():
            setattr(self, k, v)

    def save_state(self, path, ext=None):

        state = self.state_dict()
        path = beam_path(path)

        try:
            from ..data import BeamData
            has_beam_ds = True
        except ImportError:
            has_beam_ds = False

        if has_beam_ds and isinstance(state, BeamData):
            state.store(path=path, file_type=ext)
        elif has_beam_ds and (not path.suffix) and ext is None:
            state = BeamData(data=state, path=path)
            state.store()
        else:
            path.write(state, ext=ext)

    def to_path(self, path, skeleton_ext=None, state_ext=None):
        path = beam_path(path)
        path.clean()
        path.mkdir()
        skeleton_ext = skeleton_ext or '.pkl'
        state_ext = state_ext or ''
        skeleton_file = f"{Processor.skeleton_file}{skeleton_ext}"
        state_file = f"{Processor.state_file}{state_ext}"

        path.joinpath(skeleton_file).write(self)
        self.save_state(path.joinpath(state_file))

    def load_state(self, path):

        path = beam_path(path)

        try:
            from ..data import BeamData
            has_beam_ds = True
        except ImportError:
            has_beam_ds = False

        if path.is_file():
            state = path.read()
        elif has_beam_ds:
            state = BeamData.from_path(path=path)
            state = state.cache()
        else:
            raise NotImplementedError

        self.load_state_dict(state)

    def nlp(self, query, llm=None, ask_kwargs=None, **kwargs):

        from ..logger import beam_logger as logger

        if llm is None:
            llm = self.llm
        elif type(llm) is str:
            from ..resource import resource
            llm = resource(llm)

        ask_kwargs = ask_kwargs or {}

        method_list = inspect.getmembers(self, predicate=inspect.isroutine)
        method_list = [m for m in method_list if not m[0].startswith('_')]
        json_output_example = json.dumps({'method': 'method_name'})
        class_doc = inspect.getdoc(self)
        class_doc = f"{class_doc}\n" if class_doc else ""

        prompt = (f"Choose the suitable method that should be used to answer the following query:\n"
                  f"Query: {query}\n"
                  f"Class: {self.__class__.__name__}\n"
                  f"{class_doc}"
                  f"Attributes: {method_list}\n"
                  f"Return your answer as a JSON object of the following form:\n"
                  f"{json_output_example}\n"
                  f"Your answer:\n\n")

        response = llm.ask(prompt, **ask_kwargs).json
        method_name = response['method']

        if method_name not in [m[0] for m in method_list]:
            raise ValueError(f"Method {method_name} not found in the list of methods")

        logger.info(f"Using method {method_name} to answer the query")

        method = getattr(self, method_name)
        sourcecode = inspect.getsource(method)

        json_output_example = {"args": ['arg1', 'arg2'], "kwargs": {'kwarg1': 'value1', 'kwarg2': 'value2'}}

        prompt = (f"Build a suitable dictionary of arguments and keyword arguments to answer the following query:\n"
                  f"Query: {query}\n"
                  f"with the class method: {method_name} (of class {self.__class__.__name__}) with source-code:\n"
                  f"{sourcecode}\n"
                  f"Return your answer as a JSON object of the following form:\n"
                  f"{json_output_example}\n"
                  f"Your answer:\n\n")

        d = llm.ask(prompt, **ask_kwargs).json

        args = d.get('args', [])
        kwargs = d.get('kwargs', {})

        logger.info(f"Using args: {args} and kwargs: {kwargs} to answer the query")

        return method(*args, **kwargs)


class Pipeline(Processor):

    def __init__(self, hparams, *ts, track_steps=False, name=None, state=None, path=None, **kwts):

        super().__init__(hparams, name=name, state=state, path=path)
        self.track_steps = track_steps
        self.steps = {}

        self.transformers = OrderedDict()
        for i, t in enumerate(ts):
            self.transformers[i] = t

        for k, t in kwts.items():
            self.transformers[k] = t

    def transform(self, x, **kwargs):

        self.steps = []

        for i, t in self.transformers.items():

            kwargs_i = kwargs[i] if i in kwargs.keys() else {}
            x = t.transform(x, **kwargs_i)

            if self.track_steps:
                self.steps[i] = x

        return x




import docker
from docker.errors import BuildError

# Initialize Docker client
client = docker.from_env()

# Define build arguments
build_args = {
    'BASE_IMAGE': 'python:3.8-slim',
    'REQUIREMENTS_FILE': './path/to/requirements.txt',
    'ALGORITHM_DIR': './path/to/algorithm',
    'ENTRYPOINT_SCRIPT': './path/to/script.py'
}

# Path to the directory containing the Dockerfile
path_to_dockerfile = '.'

try:
    # Build the image
    image, build_logs = client.images.build(path=path_to_dockerfile, buildargs=build_args, tag='your-image-name')

    # Print build logs (optional)
    for line in build_logs:
        if 'stream' in line:
            print(line['stream'].strip())

except BuildError as e:
    print("Error building Docker image:", e)
except Exception as e:
    print("Error:", e)
finally:
    client.close()


import json
import time
from threading import Thread
import io
import websockets
import asyncio
from flask import request, jsonify, send_file

from ..serve.http_server import HTTPServer
from ..serve.server import BeamServer
from ..logger import beam_logger as logger
from ..utils import ThreadSafeDict, find_port


class AsyncServer(HTTPServer):

    def __init__(self, dispatcher, *args, asynchronous=True, postrun=None, ws_tls=False, **kwargs):

        super().__init__(dispatcher, *args, **kwargs)

        self.app.add_url_rule('/poll/<client>/', view_func=self.poll)

        self.tasks = ThreadSafeDict()
        self.ws_clients = ThreadSafeDict()
        # task_postrun.connect(self.postprocess)
        self.asynchronous = asynchronous

        if ws_tls:
            import ssl
            # Create an SSL context for wss
            self.ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            # self.ssl_context.load_cert_chain(certfile='path/to/cert.pem', keyfile='path/to/key.pem')
        else:
            self.ssl_context = None

        self.ws_application = 'ws' if not ws_tls else 'wss'
        self.ws_port = None

        if postrun is None:
            self.postrun_callback = self.postrun
        else:
            self.postrun_callback = postrun

    def run_worker(self):
        pass

    def poll(self, client):

        task_id = request.args.get('task_id')
        timeout = request.args.get('timeout', 0)
        result = self.obj.poll(task_id, timeout=timeout)

        if client == 'beam':
            io_results = io.BytesIO()
            self.dump_function(result, io_results)
            io_results.seek(0)
            return send_file(io_results, mimetype="text/plain")
        else:
            return jsonify(result)

    async def websocket_handler(self, ws):
        # Wait for the client to send its client_id

        client_id = await ws.recv()
        logger.info(f"New WebSocket client connected: {client_id}")
        self.ws_clients[client_id] = ws
        await ws.wait_closed()
        # self.ws_clients.pop(client_id)

    async def run_ws_server(self, host, port):
        logger.info("Starting WebSocket server...")
        async with websockets.serve(self.websocket_handler, host, port, ssl=self.ssl_context):
            await asyncio.Future()  # Run forever until Future is cancelled

    def run_non_blocking(self, **kwargs):
        self.run(non_blocking=True, **kwargs)

    def run(self, ws_host=None, ws_port=None, enable_websocket=True, non_blocking=False, **kwargs):

        self.run_worker()

        if enable_websocket and self.asynchronous:
            if ws_host is None:
                ws_host = "0.0.0.0"
            ws_port = find_port(port=ws_port, get_port_from_beam_port_range=True, application=self.ws_application)
            self.ws_port = ws_port
            logger.info(f"Opening a Websocket ({self.ws_application}) serve on port: {ws_port}")

            # Run the WebSocket server as an asyncio task
            # Thread(target=self.run_ws_server, args=(ws_host, ws_port)).start()
            Thread(target=lambda: asyncio.run(self.run_ws_server(ws_host, ws_port))).start()

        super().run(non_blocking=True, **kwargs)

        if not non_blocking:
            self.routine()
        else:
            Thread(target=self.routine).start()

    def routine(self):

        while True:
            try:
                sleep = True
                for task_id, task in self.tasks.items():
                    res = task['async_result']
                    if res.is_ready:
                        sleep = False
                        task_inf = self.tasks.pop(task_id)
                        self.postprocess(task_id=task_id, async_result=res, task_inf=task_inf)
                if sleep:
                    time.sleep(0.1)
            except KeyboardInterrupt:
                logger.info("Keyboard interrupt, exiting...")
                break

    @property
    def real_object(self):
        return self.obj.obj

    @property
    def metadata(self):
        return {'ws_application': self.ws_application, 'ws_port': self.ws_port}

    def query_algorithm(self, client, method, *args, **kwargs):

        if not self.asynchronous:
            return super().query_algorithm(client, method, *args, **kwargs)

        if client == 'beam':
            postrun_args = request.files['postrun_args']
            postrun_kwargs = request.files['postrun_kwargs']
            args = request.files['args']
            kwargs = request.files['kwargs']
            ws_client_id = request.files['ws_client_id']

            postrun_args = self.load_function(postrun_args)
            postrun_kwargs = self.load_function(postrun_kwargs)
            ws_client_id = self.load_function(ws_client_id)

        else:
            data = request.get_json()
            postrun_args = data.pop('postrun_args', [])
            postrun_kwargs = data.pop('postrun_kwargs', {})
            ws_client_id = data.pop('ws_client_id', None)

            args = data.pop('args', [])
            kwargs = data.pop('kwargs', {})

        if method not in ['poll']:
            async_result = BeamServer.query_algorithm(self, client, method, args, kwargs, return_raw_results=True)
            task_id = async_result.hex
            metadata = self.request_metadata(client=client, method=method)
            self.tasks[task_id] = {'metadata': metadata, 'postrun_args': postrun_args,
                                            'postrun_kwargs': postrun_kwargs, 'ws_client_id': ws_client_id,
                                            'async_result': async_result}

            if client == 'beam':
                io_results = io.BytesIO()
                self.dump_function(task_id, io_results)
                io_results.seek(0)
                return send_file(io_results, mimetype="text/plain")
            else:
                return jsonify(task_id)

        else:
            return BeamServer.query_algorithm(self, client, method, args, kwargs)

    def postprocess(self, task_id=None, task_inf=None, async_result=None):

        state = async_result.state
        args = async_result.args
        kwargs = async_result.kwargs
        retval = async_result.value

        logger.info(f"Task {task_id} finished with state {state}.")

        if task_inf is None:
            logger.warning(f"Task {task_id} not found in tasks dict")
            return

        # Send notification to the client via WebSocket
        if task_inf['ws_client_id'] is not None:
            client_id = task_inf['ws_client_id']
            ws = self.ws_clients.get(client_id)
            if ws and ws.open:
                asyncio.run(ws.send(json.dumps({"task_id": task_id, "state": state})))
            else:
                if ws:
                    asyncio.run(ws.close())
                    self.ws_clients.pop(client_id)

        self.postrun_callback(task_args=args, task_kwargs=kwargs, retval=retval, state=state, task=async_result, **task_inf)

    def postrun(self, task_args=None, task_kwargs=None, retval=None, state=None, task=None, metadata=None,
                postrun_args=None, postrun_kwargs=None, **kwargs):
        pass


class AsyncCeleryServer(AsyncServer):

    def __init__(self, obj, routes=None, name=None, asynchronous=True, broker=None, backend=None,
                 broker_username=None, broker_password=None, broker_port=None, broker_scheme=None, broker_host=None,
                 backend_username=None, backend_password=None, backend_port=None, backend_scheme=None,
                 backend_host=None, use_torch=True, batch=None, max_wait_time=1.0, max_batch_size=10,
                 tls=False, n_threads=4, application=None, n_workers=1, broker_log_level='INFO', **kwargs):

        from .celery_dispatcher import CeleryDispatcher
        from .celery_worker import CeleryWorker

        if routes is None:
            routes = []
        self.worker = CeleryWorker(obj, *routes, name=name, n_workers=n_workers, daemon=True,
                                   broker=broker, backend=backend,
                                   broker_username=broker_username, broker_password=broker_password,
                                   broker_port=broker_port,
                                   broker_scheme=broker_scheme, broker_host=broker_host,
                                   backend_username=backend_username, backend_password=backend_password,
                                   backend_port=backend_port,
                                   backend_scheme=backend_scheme,
                                   backend_host=backend_host, log_level=broker_log_level)

        predefined_attributes = {k: 'method' for k in self.worker.routes}
        dispatcher = CeleryDispatcher(name=self.worker.name, broker=broker, backend=backend,
                                           broker_username=broker_username, broker_password=broker_password,
                                           broker_port=broker_port, broker_scheme=broker_scheme,
                                           broker_host=broker_host,
                                           backend_username=backend_username, backend_password=backend_password,
                                           backend_port=backend_port, backend_scheme=backend_scheme,
                                           backend_host=backend_host, log_level=broker_log_level,
                                           asynchronous=asynchronous)

        application = application or 'distributed_celery_async'
        super().__init__(dispatcher, name=name, asynchronous=asynchronous, use_torch=use_torch, batch=batch,
                         max_wait_time=max_wait_time, max_batch_size=max_batch_size,
                         tls=tls, n_threads=n_threads, application=application,
                         predefined_attributes=predefined_attributes, **kwargs)

    def run_worker(self):
        self.worker.run()


class AsyncRayServer(AsyncServer):

    def __init__(self, obj, routes=None, name=None, asynchronous=True, address=None, port=None, use_torch=True, batch=None,
                 max_wait_time=1.0, max_batch_size=10, tls=False, n_threads=4, application=None, ray_kwargs=None,
                 **kwargs):

        from .ray_dispatcher import RayDispatcher

        if routes is None:
            routes = []

        predefined_attributes = {k: 'method' for k in routes}
        dispatcher = RayDispatcher(obj, *routes, name=name, address=address, port=port, ray_kwargs=ray_kwargs,
                                   asynchronous=asynchronous, **kwargs)

        application = application or 'distributed_ray_async'
        super().__init__(dispatcher, name=name, asynchronous=asynchronous, use_torch=use_torch, batch=batch,
                         max_wait_time=max_wait_time, max_batch_size=max_batch_size,
                         tls=tls, n_threads=n_threads, application=application,
                         predefined_attributes=predefined_attributes, **kwargs)
